import collections
from torch import nn
import torch 
from ..base import LinkPredictionLayerBase

class ElementSumLayer(LinkPredictionLayerBase):
    r"""Specific class for link prediction task.


    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    num_class : int 
               The number of node catrgoriey for classification
    hidden_size : list of int type values
                  Example for two layers's FeedforwardNN: [50, 20]
    activation: the activation function class for each fully connected layer
                Default: nn.ReLU()
                Example: nn.ReLU(),nn.Sigmoid().        

    """     
    def __init__(self, input_size, hidden_size,num_class, activation=nn.ReLU()):        
        super(ElementSumLayer, self).__init__() 
            
        #build the linear module list
        self.activation=activation
        self.ffnn_src = nn.Linear(input_size, hidden_size)
        self.ffnn_dst = nn.Linear(input_size, hidden_size)
        self.ffnn_all = nn.Linear(hidden_size, num_class)
        
    def forward(self, node_emb, edge_idx=None):
        r"""
        Forward functions to compute the logits tensor for link prediction.
      
        Parameters
        ----------
    
        node_emb : tensor [N,H]  
            N: number of nodes    
            H: length of the node embeddings
        edge_idx : a list of index of edge (represented as tuple of nodes pair indexes) that needs prediction.
            Default: 'None', doing link prediction for all pairs of nodes.
            Example: [(1,2),(1,0),(2,9)]
    
        Returns 
        -------
        logit tensor: [M, num_class] The score logits for all links that need to be preidcted.
            If edge_idx is given, the order of the predicted logits for edges is the same with that in the edge_idx
            If full prediction is select (default),the order of predicted logits are like:
                "[(0,0),(0,1),...(0,N),(1,0),(1,1),....(N,N)]"
        """ 
        if edge_idx == None:
            #get the index list for all the node pairs
             num_node=node_emb.shape[0]
             node_idx_list=[idx for idx in range(num_node)]
             src_idx=torch.tensor(node_idx_list).view(-1,1).repeat(1,num_node).view(-1)
             dst_idx=torch.tensor(node_idx_list).view(1,-1).repeat(num_node,1).view(-1)
        else:
            #get the index list for required pairs of nodes
            src_idx=torch.tensor([tuple_idx[0] for tuple_idx in edge_idx])    
            dst_idx=torch.tensor([tuple_idx[1] for tuple_idx in edge_idx])    

        scr_emb = self.ffnn_src(node_emb[src_idx, :])  # input the source node embeddings into ffnn
        dst_emb = self.ffnn_dst(node_emb[dst_idx, :])  # input the destinate node embeddings into ffnn

        return self.ffnn_all(self.activation(scr_emb+dst_emb))






import torch
import dgl
import torch.nn as nn
from dgl.nn import GatedGraphConv
import dgl.function as fn
from torch.nn import init
from .base import GNNLayerBase, GNNBase
from ...data.data import GraphData


class UndirectedGGNNLayerConv(GNNLayerBase):
    r"""
    Gated Graph Convolution layer from paper `Gated Graph Sequence
    Neural Networks <https://arxiv.org/pdf/1511.05493.pdf>`__.
    .. math::
       h_{i}^{0} & = [ x_i \| \mathbf{0} ]
       a_{i}^{t} & = \sum_{j\in\mathcal{N}(i)} W_{e_{ij}} h_{j}^{t}
       h_{i}^{t+1} & = \mathrm{GRU}(a_{i}^{t}, h_{i}^{t})
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    num_layers: int
        Number of GGNN layers. Default: 1.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.

    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    num_layers: int
        Number of GGNN layers. Default: 1.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.
    """

    def __init__(self,
                 input_size,
                 output_size,
                 num_layers,
                 n_etypes,
                 bias=True):
        super(UndirectedGGNNLayerConv, self).__init__()
        self._in_feats = input_size
        self._out_feats = output_size
        self._num_layers = num_layers
        self._n_etypes = n_etypes
        self.linears = nn.ModuleList(
            [nn.Linear(output_size, output_size) for _ in range(n_etypes)]
        )
        self.gru = nn.GRUCell(output_size, output_size, bias=bias)
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = init.calculate_gain('relu')
        self.gru.reset_parameters()
        for linear in self.linears:
            init.xavier_normal_(linear.weight, gain=gain)
            init.zeros_(linear.bias)

    def forward(self, graph, feat, etypes=None, edge_weight=None):
        """Compute Gated Graph Convolution layer.
        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat : torch.Tensor
            The input feature of shape :math:`(N, D_{in})` where :math:`N`
            is the number of nodes of the graph and :math:`D_{in}` is the
            input feature size.
        etypes : torch.LongTensor
            The edge type tensor of shape :math:`(E,)` where :math:`E` is
            the number of edges of the graph.
        edge_weight: torch.Tensor
            The shape of edge_weight is :math:`(N_E, 1)`. N_E is the number of edges in graph.
        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, D_{out})` where
            :math:`D_{out}` is size of output feature.
        """
        if dgl.__version__ < '0.5':
            assert graph.is_homograph(), "not a homograph; convert it with to_homo and pass in the edge type as argument"
        else:
            assert graph.is_homogeneous, "not a homograph; convert it with to_homo and pass in the edge type as argument"
        graph = graph.local_var()
        zero_pad = feat.new_zeros((feat.shape[0], self._out_feats - feat.shape[1]))
        feat = torch.cat([feat, zero_pad], -1)

        for _ in range(self._num_layers):
            graph.ndata['h'] = feat
            for i in range(self._n_etypes):
                eids = (etypes == i).nonzero().view(-1)
                if len(eids) > 0:
                    if edge_weight is None:
                        graph.apply_edges(
                            lambda edges: {'W_e*h': self.linears[i](edges.src['h'])},
                            eids
                        )
                    else:
                        graph.apply_edges(
                            lambda edges: {'W_e*h': self.linears[i](edges.src['h']) * edge_weight.view(-1).unsqueeze(1)},
                            eids
                        )
            graph.update_all(fn.copy_e('W_e*h', 'm'), fn.sum('m', 'a'))
            a = graph.ndata.pop('a')  # (N, D)
            feat = self.gru(a, feat)
        return feat


class BiFuseGGNNLayerConv(GNNLayerBase):
    r"""
    Fuse aggregated embeddings from both incoming and outgoing
    directions before updating node embeddings.
    .. math::
       h_{i}^{0} & = [ x_i \| \mathbf{0} ]
       a_{i, \vdash}^{t} & = \sum_{j\in\mathcal{N}_{\vdash }(i)}
       W_{\vdash} h_{j}^{t}
       a_{i, \dashv}^{t} & = \sum_{j\in\mathcal{N}_{\dashv }(i)}
       W_{\dashv} h_{j}^{t}
       e_{i}^{t} &= \sigma (W_{f}[a_{i, \vdash}^{t};a_{i, \dashv}^{t};
       a_{i, \vdash}^{t}*a_{i, \dashv}^{t};
       a_{i, \vdash}^{t}-a_{i, \dashv}^{t}])
       h_{i}^{t+1} & = \mathrm{GRU}(e_{i}^{t}, h_{i}^{t})
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.
    """

    def __init__(self, input_size, output_size, n_etypes=1, bias=True):
        super(BiFuseGGNNLayerConv, self).__init__()
        self._input_size = input_size
        self._output_size = output_size
        # self._num_layers = num_layers
        self._n_etypes = n_etypes

        self.linears_in = nn.ModuleList(
            [nn.Linear(output_size, output_size) for _ in range(n_etypes)]
        )

        self.linears_out = nn.ModuleList(
            [nn.Linear(output_size, output_size) for _ in range(n_etypes)]
        )

        self.gru = nn.GRUCell(output_size, output_size, bias=bias)
        self.fuse_linear = nn.Linear(4 * output_size, output_size, bias=True)
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        self.gru.reset_parameters()
        nn.init.xavier_normal_(self.fuse_linear.weight, gain=gain)

        for linear in self.linears_in:
            nn.init.xavier_normal_(linear.weight, gain=gain)
            nn.init.zeros_(linear.bias)

        for linear in self.linears_out:
            nn.init.xavier_normal_(linear.weight, gain=gain)
            nn.init.zeros_(linear.bias)

    def forward(self, graph, node_feats, etypes=None, edge_weight=None):
        """
        Parameters
        ----------
        graph: dgl.DGLGraph
        node_feats: torch.Tensor
            The shape of node_feats is :math:`(N, D_{in})`.
        edge_weight: torch.Tensor
            The shape of edge_weight is :math:`(N_E, 1)`. N_E is the number of edges in graph.
        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, D_{out})` where
            :math:`D_{out}` is size of output feature.
        """
        feat_in, feat_out = node_feats  # feat_in == feat_out
        # etypes = torch.LongTensor([0] * graph.number_of_edges())  # [B, E]. E is the number of edges.

        # forward aggregation
        graph_in = graph
        graph_in = graph_in.local_var()
        graph_in.ndata['h'] = feat_in
        for i in range(self._n_etypes):
            eids = (etypes == i).nonzero().view(-1)
            if len(eids) > 0:
                if edge_weight is None:
                    graph_in.apply_edges(
                        lambda edges: {'W_e*h': self.linears_in[i](edges.src['h'])},
                        eids
                    )
                else:
                    assert isinstance(edge_weight, tuple)
                    graph_in.apply_edges(
                        lambda edges: {'W_e*h': self.linears_in[i](edges.src['h']) * edge_weight[0].view(-1).unsqueeze(1)},
                        eids
                    )
        graph_in.update_all(fn.copy_e('W_e*h', 'm'), fn.sum('m', 'a'))
        agg_in = graph_in.ndata.pop('a')  # (N, D)

        # backward aggregation
        graph_out = graph.reverse()
        graph_out = graph_out.local_var()
        graph_out.ndata['h'] = feat_out
        for i in range(self._n_etypes):
            eids = (etypes == i).nonzero().view(-1)
            if len(eids) > 0:
                if edge_weight is None:
                    graph_out.apply_edges(
                        lambda edges: {'W_e*h': self.linears_out[i](edges.src['h'])},
                        eids
                    )
                else:
                    assert isinstance(edge_weight, tuple)
                    graph_out.apply_edges(
                        lambda edges: {'W_e*h': self.linears_out[i](edges.src['h']) * edge_weight[1].view(-1).unsqueeze(1)},
                        eids
                    )
        graph_out.update_all(fn.copy_e('W_e*h', 'm'), fn.sum('m', 'a'))
        agg_out = graph_out.ndata.pop('a')  # (N, D)

        # fuse
        fuse_vector = torch.cat(
            [agg_in, agg_out, agg_in * agg_out, agg_in - agg_out], dim=-1)
        fuse_gate_vector = torch.sigmoid(self.fuse_linear(fuse_vector))
        emb_fused = fuse_gate_vector * agg_in + (1 - fuse_gate_vector) * agg_out

        # update
        rst = self.gru(emb_fused, feat_in)

        return [rst, rst]


class BiSepGGNNLayerConv(GNNLayerBase):
    r"""
    Compute node embeddings for incoming and outgoing directions
    separately, and then concatenate the two output node embeddings
    after the final layer.
    .. math::
       h_{i}^{0} & = [ x_i \| \mathbf{0} ]
       a_{i, \vdash}^{t} & = \sum_{j\in\mathcal{N}_{\vdash }(i)}
       W_{\vdash} h_{j, \vdash}^{t}
       a_{i, \dashv}^{t} & = \sum_{j\in\mathcal{N}_{\dashv }(i)}
       W_{\dashv} h_{j, \dashv}^{t}
       h_{i, \vdash}^{t+1} & = \mathrm{GRU}_{\vdash}(a_{i, \vdash}^{t},
       h_{i, \vdash}^{t})
       h_{i, \dashv}^{t+1} & = \mathrm{GRU}_{\dashv}(a_{i, \dashv}^{t},
       h_{i, \dashv}^{t})
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    n_etypes: int
        Number of edge types. Default: 1.
    bias: bool
        If True, adds a learnable bias to the output. Default: True.
    """

    def __init__(self, input_size, output_size, n_etypes=1, bias=True):
        super(BiSepGGNNLayerConv, self).__init__()
        self._input_size = input_size
        self._output_size = output_size
        # self._num_layers = num_layers
        self._n_etypes = n_etypes

        self.linears_in = nn.ModuleList(
            [nn.Linear(output_size, output_size) for _ in range(n_etypes)]
        )

        self.linears_out = nn.ModuleList(
            [nn.Linear(output_size, output_size) for _ in range(n_etypes)]
        )

        # self.update_in = nn.Linear(input_size + output_size, output_size, bias=True)
        # self.update_out = nn.Linear(input_size + output_size, output_size, bias=True)

        self.gru_in = nn.GRUCell(output_size, output_size, bias=bias)
        self.gru_out = nn.GRUCell(output_size, output_size, bias=bias)
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        self.gru_in.reset_parameters()
        self.gru_out.reset_parameters()

        for linear in self.linears_in:
            nn.init.xavier_normal_(linear.weight, gain=gain)
            nn.init.zeros_(linear.bias)

        for linear in self.linears_out:
            nn.init.xavier_normal_(linear.weight, gain=gain)
            nn.init.zeros_(linear.bias)

        # nn.init.xavier_normal_(self.update_in.weight, gain=gain)
        # nn.init.xavier_normal_(self.update_out.weight, gain=gain)

    def forward(self, graph, node_feats, etypes=None, edge_weight=None):
        feat_in, feat_out = node_feats
        # etypes = torch.LongTensor([0] * graph.number_of_edges())  # [B, E]. E is the number of edges.

        graph_in = graph
        graph_in = graph_in.local_var()
        graph_in.ndata['h'] = feat_in
        for i in range(self._n_etypes):
            eids = (etypes == i).nonzero().view(-1)
            if len(eids) > 0:
                if edge_weight is None:
                    graph_in.apply_edges(
                        lambda edges: {'W_e*h': self.linears_in[i](edges.src['h'])},
                        eids
                    )
                else:
                    assert isinstance(edge_weight, tuple)
                    graph_in.apply_edges(
                        lambda edges: {'W_e*h': self.linears_in[i](edges.src['h']) * edge_weight[0].view(-1).unsqueeze(1)},
                        eids
                    )
        graph_in.update_all(fn.copy_e('W_e*h', 'm'), fn.sum('m', 'a'))
        a_in = graph_in.ndata.pop('a')  # (N, D)
        emb_in = self.gru_in(a_in, feat_in)

        graph_out = graph.reverse()
        graph_out = graph_out.local_var()
        graph_out.ndata['h'] = feat_out
        for i in range(self._n_etypes):
            eids = (etypes == i).nonzero().view(-1)
            if len(eids) > 0:
                if type(edge_weight) == type(None):
                    graph_out.apply_edges(
                        lambda edges: {'W_e*h': self.linears_out[i](edges.src['h'])},
                        eids
                    )
                else:
                    assert isinstance(edge_weight, tuple)
                    graph_out.apply_edges(
                        lambda edges: {'W_e*h': self.linears_out[i](edges.src['h']) * edge_weight[1].view(-1).unsqueeze(1)},
                        eids
                    )
        graph_out.update_all(fn.copy_e('W_e*h', 'm'), fn.sum('m', 'a'))
        a_out = graph_out.ndata.pop('a')  # (N, D)
        emb_out = self.gru_out(a_out, feat_out)

        return [emb_in, emb_out]


class GGNNLayer(GNNLayerBase):
    r"""A unified wrapper for Gated Graph Convolution layer
    from paper `Gated Graph Sequence Neural Networks
    <https://arxiv.org/pdf/1511.05493.pdf>`__.
    Support both undirected (i.e., regular) and bidirectional
    (i.e., `bi_sep` and `bi_fuse`) versions.
    Attributes
    ----------
    input_size: int
        Input feature size.
    output_size: int
        Output feature size.
    direction_option: str
        The direction option of GGNN ('undirected', 'bi_sep' or 'bi_fuse'). (Default: 'bi_fuse')
    num_layers: int
        Number of GGNN layers. `num_layers` is set to any integer if the direction_option is 'undirected'.
        If the direction_option is 'bi_sep' or 'bi_fuse', `num_layers` will be set to 1.
    n_etypes: int
        Number of edge types. `n_etypes` can be set to any integer if the direction_option is 'undirected'.
        If the direction_option is 'bi_sep' or 'bi_fuse', `n_etypes` will be set to 1.
    bias: bool
        If True, adds a learnable bias to the output. (Default: True)
    """

    def __init__(self, input_size, output_size, direction_option='bi_fuse', num_layers=1, n_etypes=1, bias=True):
        super(GGNNLayer, self).__init__()
        if direction_option == 'undirected':
            self.model = UndirectedGGNNLayerConv(input_size, output_size, num_layers=num_layers, n_etypes=n_etypes, bias=bias)
        elif direction_option == 'bi_sep':
            self.model = BiSepGGNNLayerConv(input_size, output_size, n_etypes, bias=bias)
        elif direction_option == 'bi_fuse':
            self.model = BiFuseGGNNLayerConv(input_size, output_size, n_etypes, bias=bias)
        else:
            raise RuntimeError('Unknown `bidirection` value: {}'.format(direction_option))

    def forward(self, graph, node_feats, etypes=None, edge_weight=None):
        """
        Parameters
        ----------
        graph: dgl.DGLGraph
        node_feats: torch.Tensor
            The shape of node_feats is :math:`(N, D_{in})`.
        Returns
        -------
        torch.Tensor
        """
        if etypes is None:
            etypes = torch.tensor([0] * graph.number_of_edges(), dtype=torch.long).to(node_feats.device)
        return self.model(graph, node_feats, etypes, edge_weight)


class GGNN(GNNBase):
    r"""
    Multi-layered `Gated Graph Sequence Neural Networks
    <https://arxiv.org/pdf/1511.05493.pdf>`__.
    Support both undirected (i.e., regular) and bidirectional
    (i.e., `bi_sep` and `bi_fuse`) versions.
    Attributes
    ----------
    num_layers: int
        Number of GGNN layers.
    input_size: int
        Input feature size.
    hidden_size: int
        Should be equal to output_size.
    output_size: int
        Output feature size.
    direction_option: str
        The direction option of GGNN ('undirected', 'bi_sep' or 'bi_fuse'). (Default: 'bi_fuse')
    n_etypes: int
        Number of edge types. n_etypes can be set to any integer if the direction_option is 'undirected'.
        If the direction_option is 'bi_sep' or 'bi_fuse', n_etypes will be set to 1.
    bias: bool
        If True, adds a learnable bias to the output. (Default: True)
    """

    def __init__(self, num_layers, input_size, hidden_size, output_size, feat_drop=0.,
                 direction_option='bi_fuse', n_etypes=1, bias=True, use_edge_weight=False):
        super(GGNN, self).__init__()
        self.num_layers = num_layers
        self.direction_option = direction_option
        self.input_size = input_size
        self.output_size = output_size
        self.feat_drop = nn.Dropout(feat_drop)
        self.use_edge_weight = use_edge_weight
        self.n_etypes = n_etypes

        assert self.output_size >= self.input_size
        assert self.output_size == hidden_size

        if self.direction_option == 'undirected':
            self.models = GGNNLayer(input_size, output_size, direction_option, num_layers=num_layers, n_etypes=n_etypes,
                                    bias=bias)
        else:
            self.models = GGNNLayer(input_size, output_size, direction_option, n_etypes=n_etypes, bias=bias)

    def forward(self, graph: GraphData):
        r"""
        Use GGNN compute node embeddings.
        Parameters
        ----------
        graph: GraphData.
            The initial node features are stored in the node feature field
            named `node_feat`.
        Returns
        -------
        input_graph: GraphData.
            The computed node embedding tensors are stored in the node feature field
            named `node_emb`.
        """
        if self.n_etypes==1:
            graph.edge_features['etype'] = torch.tensor([0] * graph.get_edge_num(), dtype=torch.long, device=graph.device)

        node_feats = graph.node_features['node_feat']
        etypes = graph.edge_features['etype']
        if self.use_edge_weight:
            edge_weight = graph.edge_features['edge_weight']
            if self.direction_option == 'bi_fuse' or self.direction_option == 'bi_sep':
                reverse_edge_weight = graph.edge_features['reverse_edge_weight']
                edge_weight = (edge_weight, reverse_edge_weight)
        else:
            edge_weight = None

        dgl_graph = graph.to_dgl()

        if self.direction_option == 'undirected':
            node_embs = self.models(dgl_graph, node_feats, etypes, edge_weight)
        else:
            assert node_feats.shape[1] == self.input_size

            zero_pad = node_feats.new_zeros((node_feats.shape[0], self.output_size - node_feats.shape[1]))
            node_feats = torch.cat([node_feats, zero_pad], -1)

            feat_in = node_feats
            feat_out = node_feats

            for i in range(self.num_layers):
                feat_in = self.feat_drop(feat_in)
                feat_out = self.feat_drop(feat_out)
                h = self.models(dgl_graph, (feat_in, feat_out), etypes, edge_weight)
                feat_in = h[0]
                feat_out = h[1]

            if self.direction_option == 'bi_sep':
                node_embs = torch.cat([feat_in, feat_out], dim=-1)
            elif self.direction_option == 'bi_fuse':
                node_embs = feat_in
            else:
                raise RuntimeError('Unknown `bidirection` value: {}'.format(self.direction_option))

        graph.node_features['node_emb'] = node_embs

        return graph


import os
import time
import datetime
import copy
import argparse
from argparse import Namespace
import yaml

import numpy as np
from scipy import sparse
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.optim.lr_scheduler import ReduceLROnPlateau
import torch.backends.cudnn as cudnn

from graph4nlp.pytorch.datasets.squad import SQuADDataset
from graph4nlp.pytorch.data.data import from_batch
from graph4nlp.pytorch.modules.graph_construction import *
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import RNNEmbedding, WordEmbedding
from graph4nlp.pytorch.models.graph2seq import Graph2Seq
from graph4nlp.pytorch.modules.utils.generic_utils import grid, to_cuda, dropout_fn, sparse_mx_to_torch_sparse_tensor, EarlyStopping
from graph4nlp.pytorch.modules.config import get_basic_args
from graph4nlp.pytorch.models.graph2seq_loss import Graph2SeqLoss
from graph4nlp.pytorch.modules.utils.copy_utils import prepare_ext_vocab
from graph4nlp.pytorch.modules.evaluation import BLEU, METEOR, ROUGE
from graph4nlp.pytorch.modules.utils.logger import Logger
from graph4nlp.pytorch.modules.utils import constants as Constants
from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals_no_size
from graph4nlp.pytorch.modules.prediction.generation.decoder_strategy import DecoderStrategy
from graph4nlp.pytorch.modules.utils.config_utils import update_values, get_yaml_config
import multiprocessing
import torch.multiprocessing
# torch.multiprocessing.set_sharing_strategy('file_system')
# multiprocessing.set_start_method("spawn", force=True)


class QGModel(nn.Module):
    def __init__(self, vocab, config):
        super(QGModel, self).__init__()
        self.config = config
        self.vocab = vocab
        self.use_coverage = self.config['decoder_args']['rnn_decoder_share']['use_coverage']

        # build Graph2Seq model
        self.g2s = Graph2Seq.from_args(config, self.vocab, config['device'])

        if 'w2v' in self.g2s.graph_topology.embedding_layer.word_emb_layers:
            self.word_emb = self.g2s.graph_topology.embedding_layer.word_emb_layers['w2v'].word_emb_layer
        else:
            self.word_emb = WordEmbedding(
                            self.vocab.in_word_vocab.embeddings.shape[0],
                            self.vocab.in_word_vocab.embeddings.shape[1],
                            pretrained_word_emb=self.vocab.in_word_vocab.embeddings,
                            fix_emb=config['graph_construction_args']['node_embedding']['fix_word_emb'],
                            device=config['device']).word_emb_layer

        answer_feat_size = self.vocab.in_word_vocab.embeddings.shape[1]
        if 'node_edge_bert' in self.g2s.graph_topology.embedding_layer.word_emb_layers:
            self.bert_encoder = self.g2s.graph_topology.embedding_layer.word_emb_layers['node_edge_bert']
            answer_feat_size += self.bert_encoder.bert_model.config.hidden_size
        elif 'seq_bert' in self.g2s.graph_topology.embedding_layer.word_emb_layers:
            self.bert_encoder = self.g2s.graph_topology.embedding_layer.word_emb_layers['seq_bert']
            answer_feat_size += self.bert_encoder.bert_model.config.hidden_size
        else:
            self.bert_encoder = None


        self.ans_rnn_encoder = RNNEmbedding(
                                    answer_feat_size,
                                    config['num_hidden'],
                                    bidirectional=True,
                                    num_layers=1,
                                    rnn_type='lstm',
                                    dropout=config['enc_rnn_dropout'],
                                    device=config['device'])

        # soft-alignment between context and answer
        self.ctx2ans_attn = Context2AnswerAttention(config['num_hidden'], config['num_hidden'])
        self.fuse_ctx_ans = nn.Linear(2 * config['num_hidden'], config['num_hidden'], bias=False)

        self.loss_calc = Graph2SeqLoss(ignore_index=self.vocab.out_word_vocab.PAD,
                                  use_coverage=self.use_coverage, coverage_weight=config['coverage_loss_ratio'])

    def encode_init_node_feature(self, data):
        num_graph_nodes = []
        for g in data['graph_data']:
            num_graph_nodes.append(g.get_node_num())

        # graph embedding construction
        batch_gd = self.g2s.graph_topology(data['graph_data'])

        # answer alignment
        answer_feat = self.word_emb(data['input_tensor2'])
        answer_feat = dropout_fn(answer_feat, self.config['word_dropout'], shared_axes=[-2], training=self.training)

        if self.bert_encoder is not None:
            answer_bert_feat = self.bert_encoder(data['input_text2'])
            answer_feat = torch.cat([answer_feat, answer_bert_feat], -1)

        answer_feat = self.ans_rnn_encoder(answer_feat, data['input_length2'])[0]
        new_node_feat = self.answer_alignment(batch_gd.node_features['node_feat'], answer_feat, answer_feat, num_graph_nodes, data['input_length2'])
        batch_gd.node_features['node_feat'] = new_node_feat

        return batch_gd

    def forward(self, data, oov_dict=None, require_loss=True):
        batch_gd = self.encode_init_node_feature(data)
        prob, enc_attn_weights, coverage_vectors = self.g2s.encoder_decoder(batch_gd, data['graph_data'], oov_dict=oov_dict, tgt_seq=data['tgt_tensor'])

        if require_loss:
            tgt = data['tgt_tensor']
            min_length = min(prob.shape[1], tgt.shape[1])
            prob = prob[:, :min_length, :]
            tgt = tgt[:, :min_length]
            loss = self.loss_calc(prob, label=tgt, enc_attn_weights=enc_attn_weights, coverage_vectors=coverage_vectors)
            return prob, loss * min_length / 2
        else:
            return prob

    def answer_alignment(self, node_feat, answer_feat, out_answer_feat, num_nodes, answer_length):
        assert len(num_nodes) == len(answer_length)
        mask = []
        for i in range(len(num_nodes)): # batch
            tmp_mask = np.ones((num_nodes[i], answer_feat.shape[1]))
            tmp_mask[:, answer_length[i]:] = 0
            mask.append(sparse.coo_matrix(tmp_mask))

        mask = sparse.block_diag(mask)
        mask = to_cuda(sparse_mx_to_torch_sparse_tensor(mask).to_dense(), self.config['device'])

        answer_feat = answer_feat.reshape(-1, answer_feat.shape[-1])
        out_answer_feat = out_answer_feat.reshape(-1, out_answer_feat.shape[-1])
        ctx_aware_ans_feat = self.ctx2ans_attn(node_feat, answer_feat, out_answer_feat, mask)
        new_node_feat = self.fuse_ctx_ans(torch.cat([node_feat, ctx_aware_ans_feat], -1))

        return new_node_feat


class Context2AnswerAttention(nn.Module):
    def __init__(self, dim, hidden_size):
        super(Context2AnswerAttention, self).__init__()
        self.linear_sim = nn.Linear(dim, hidden_size, bias=False)

    def forward(self, context, answers, out_answers, mask=None):
        """
        Parameters
        :context, (L, dim)
        :answers, (N, dim)
        :out_answers, (N, dim)
        :mask, (L, N)

        Returns
        :ques_emb, (L, dim)
        """
        context_fc = torch.relu(self.linear_sim(context))
        questions_fc = torch.relu(self.linear_sim(answers))

        # shape: (L, N)
        attention = torch.matmul(context_fc, questions_fc.transpose(-1, -2))
        if mask is not None:
            attention = attention.masked_fill_(~mask.bool(), -Constants.INF)
        prob = torch.softmax(attention, dim=-1)
        # shape: (L, dim)
        emb = torch.matmul(prob, out_answers)

        return emb


class ModelHandler:
    def __init__(self, config):
        super(ModelHandler, self).__init__()
        self.config = config
        self.use_copy = self.config['decoder_args']['rnn_decoder_share']['use_copy']
        self.use_coverage = self.config['decoder_args']['rnn_decoder_share']['use_coverage']
        self.logger = Logger(config['out_dir'], config={k:v for k, v in config.items() if k != 'device'}, overwrite=True)
        self.logger.write(config['out_dir'])
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        if self.config['graph_construction_args']["graph_construction_share"]["graph_type"] == "dependency":
            topology_builder = DependencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.config['graph_construction_args']["graph_construction_share"]["graph_type"] == "constituency":
            topology_builder = ConstituencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.config['graph_construction_args']["graph_construction_share"]["graph_type"] == "node_emb":
            topology_builder = NodeEmbeddingBasedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_topology_builder = None
        elif self.config['graph_construction_args']["graph_construction_share"]["graph_type"] == "node_emb_refined":
            topology_builder = NodeEmbeddingBasedRefinedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_graph_type = self.config['graph_construction_args'].graph_construction_private.dynamic_init_graph_type
            if dynamic_init_graph_type is None or dynamic_init_graph_type == 'line':
                dynamic_init_topology_builder = None
            elif dynamic_init_graph_type == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif dynamic_init_graph_type == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError('Define your own dynamic_init_topology_builder')
        else:
            raise NotImplementedError("Define your topology builder.")

        dataset = SQuADDataset(root_dir=self.config['graph_construction_args']['graph_construction_share']['root_dir'],
                              pretrained_word_emb_file=self.config['pre_word_emb_file'],
                              merge_strategy=self.config['graph_construction_args']['graph_construction_private']['merge_strategy'],
                              edge_strategy=self.config['graph_construction_args']["graph_construction_private"]['edge_strategy'],
                              max_word_vocab_size=self.config['top_word_vocab'],
                              min_word_vocab_freq=self.config['min_word_freq'],
                              word_emb_size=self.config['word_emb_size'],
                              share_vocab=self.config['share_vocab'],
                              seed=self.config['seed'],
                              graph_type=graph_type,
                              topology_builder=topology_builder,
                              topology_subdir=self.config['graph_construction_args']['graph_construction_share']['topology_subdir'],
                              dynamic_graph_type=self.config['graph_construction_args']['graph_construction_share']['graph_type'],
                              dynamic_init_topology_builder=dynamic_init_topology_builder,
                              dynamic_init_topology_aux_args={'dummy_param': 0},
                              thread_number=4,
                              port=9000,
                              timeout=15000)

        # TODO: use small ratio of the data (Test only)
        dataset.train = dataset.train[:self.config['n_samples']]
        dataset.val = dataset.val[:self.config['n_samples']]
        dataset.test = dataset.test[:self.config['n_samples']]

        self.train_dataloader = DataLoader(dataset.train, batch_size=self.config['batch_size'], shuffle=True,
                                           num_workers=self.config['num_workers'],
                                           collate_fn=dataset.collate_fn)
        self.val_dataloader = DataLoader(dataset.val, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model
        self.num_train = len(dataset.train)
        self.num_val = len(dataset.val)
        self.num_test = len(dataset.test)
        print('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))
        self.logger.write('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))

    def _build_model(self):
        self.model = QGModel(self.vocab, self.config).to(self.config['device'])

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=self.config['lr'])
        self.stopper = EarlyStopping(os.path.join(self.config['out_dir'], Constants._SAVED_WEIGHTS_FILE), patience=self.config['patience'])
        self.scheduler = ReduceLROnPlateau(self.optimizer, mode='max', factor=self.config['lr_reduce_factor'], \
            patience=self.config['lr_patience'], verbose=True)

    def _build_evaluation(self):
        self.metrics = {'BLEU': BLEU(n_grams=[1, 2, 3, 4]),
                        'METEOR': METEOR(),
                        'ROUGE': ROUGE()}

    def train(self):
        dur = []
        for epoch in range(self.config['epochs']):
            self.model.train()
            train_loss = []
            t0 = time.time()
            for i, data in enumerate(self.train_dataloader):
                data = all_to_cuda(data, self.config['device'])

                oov_dict = None
                if self.use_copy:
                    oov_dict, tgt = prepare_ext_vocab(data['graph_data'], self.vocab, gt_str=data['tgt_text'],
                                                        device=self.config['device'])
                    data['tgt_tensor'] = tgt

                logits, loss = self.model(data, oov_dict=oov_dict, require_loss=True)
                self.optimizer.zero_grad()
                loss.backward()
                if self.config.get('grad_clipping', None) not in (None, 0):
                    # Clip gradients
                    parameters = [p for p in self.model.parameters() if p.requires_grad]
                    # if self.config['use_bert'] and self.config.get('finetune_bert', None):
                    #     parameters += [p for p in self.config['bert_model'].parameters() if p.requires_grad]

                    torch.nn.utils.clip_grad_norm_(parameters, self.config['grad_clipping'])

                self.optimizer.step()
                train_loss.append(loss.item())

                pred = torch.max(logits, dim=-1)[1].cpu()
                dur.append(time.time() - t0)

            val_scores = self.evaluate(self.val_dataloader)
            self.scheduler.step(val_scores[self.config['early_stop_metric']])
            format_str = 'Epoch: [{} / {}] | Time: {:.2f}s | Loss: {:.4f} | Val scores:'.format(epoch + 1, self.config['epochs'], np.mean(dur), np.mean(train_loss))
            format_str += self.metric_to_str(val_scores)
            print(format_str)
            self.logger.write(format_str)

            if self.stopper.step(val_scores[self.config['early_stop_metric']], self.model):
                break

        return self.stopper.best_score

    def evaluate(self, dataloader):
        self.model.eval()
        with torch.no_grad():
            pred_collect = []
            gt_collect = []
            for i, data in enumerate(dataloader):
                data = all_to_cuda(data, self.config['device'])

                if self.use_copy:
                    oov_dict = prepare_ext_vocab(data['graph_data'], self.vocab, device=self.config['device'])
                    ref_dict = oov_dict
                else:
                    oov_dict = None
                    ref_dict = self.vocab.out_word_vocab

                prob = self.model(data, oov_dict=oov_dict, require_loss=False)
                pred = prob.argmax(dim=-1)

                pred_str = wordid2str(pred.detach().cpu(), ref_dict)
                pred_collect.extend(pred_str)
                gt_collect.extend(data['tgt_text'])

            scores = self.evaluate_predictions(gt_collect, pred_collect)

            return scores

    def translate(self, dataloader):
        self.model.eval()
        with torch.no_grad():
            pred_collect = []
            gt_collect = []
            for i, data in enumerate(dataloader):
                data = all_to_cuda(data, self.config['device'])
                if self.use_copy:
                    oov_dict = prepare_ext_vocab(data['graph_data'], self.vocab, device=self.config['device'])
                    ref_dict = oov_dict
                else:
                    oov_dict = None
                    ref_dict = self.vocab.out_word_vocab

                batch_gd = self.model.encode_init_node_feature(data)
                prob = self.model.g2s.encoder_decoder_beam_search(batch_gd, data['graph_data'], self.config['beam_size'], topk=1, oov_dict=oov_dict)

                pred_ids = torch.zeros(len(prob), self.config['decoder_args']['rnn_decoder_private']['max_decoder_step']).fill_(ref_dict.EOS).to(self.config['device']).int()
                for i, item in enumerate(prob):
                    item = item[0]
                    seq = [j.view(1, 1) for j in item]
                    seq = torch.cat(seq, dim=1)
                    pred_ids[i, :seq.shape[1]] = seq

                pred_str = wordid2str(pred_ids.detach().cpu(), ref_dict)

                pred_collect.extend(pred_str)
                gt_collect.extend(data['tgt_text'])

            scores = self.evaluate_predictions(gt_collect, pred_collect)

            return scores

    def test(self):
        # restored best saved model
        self.stopper.load_checkpoint(self.model)

        t0 = time.time()
        scores = self.translate(self.test_dataloader)
        dur = time.time() - t0
        format_str = 'Test examples: {} | Time: {:.2f}s |  Test scores:'.format(self.num_test, dur)
        format_str += self.metric_to_str(scores)
        print(format_str)
        self.logger.write(format_str)

        return scores

    def evaluate_predictions(self, ground_truth, predict):
        output = {}
        for name, scorer in self.metrics.items():
            score = scorer.calculate_scores(ground_truth=ground_truth, predict=predict)
            if name.upper() == 'BLEU':
                for i in range(len(score[0])):
                    output['BLEU_{}'.format(i + 1)] = score[0][i]
            else:
                output[name] = score[0]

        return output

    def metric_to_str(self, metrics):
        format_str = ''
        for k in metrics:
            format_str += ' {} = {:0.5f},'.format(k.upper(), metrics[k])

        return format_str[:-1]


def main(config):
    # configure
    np.random.seed(config['seed'])
    torch.manual_seed(config['seed'])

    if not config['no_cuda'] and torch.cuda.is_available():
        print('[ Using CUDA ]')
        config['device'] = torch.device('cuda' if config['gpu'] < 0 else 'cuda:%d' % config['gpu'])
        cudnn.benchmark = True
        torch.cuda.manual_seed(config['seed'])
    else:
        config['device'] = torch.device('cpu')

    ts = datetime.datetime.now().timestamp()
    # config['out_dir'] += '_{}'.format(ts)
    print('\n' + config['out_dir'])

    runner = ModelHandler(config)
    t0 = time.time()

    val_score = runner.train()
    test_scores = runner.test()

    # print('Removed best saved model file to save disk space')
    # os.remove(runner.stopper.save_model_path)
    runtime = time.time() - t0
    print('Total runtime: {:.2f}s'.format(time.time() - t0))
    runner.logger.write('Total runtime: {:.2f}s\n'.format(runtime))
    runner.logger.close()

    return val_score, test_scores

def wordid2str(word_ids, vocab):
    ret = []
    assert len(word_ids.shape) == 2, print(word_ids.shape)
    for i in range(word_ids.shape[0]):
        id_list = word_ids[i, :]
        ret_inst = []
        for j in range(id_list.shape[0]):
            if id_list[j] == vocab.EOS:
                break
            token = vocab.getWord(id_list[j])
            ret_inst.append(token)
        ret.append(" ".join(ret_inst))
    return ret

def all_to_cuda(data, device=None):
    if isinstance(data, torch.Tensor):
        data = to_cuda(data, device)
    elif isinstance(data, (list, dict)):
        keys = range(len(data)) if isinstance(data, list) else data.keys()
        for k in keys:
            if isinstance(data[k], torch.Tensor):
                data[k] = to_cuda(data[k], device)

    return data

################################################################################
# ArgParse and Helper Functions #
################################################################################
def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-task_config', '--task_config', required=True, type=str, help='path to the config file')
    parser.add_argument('-g2s_config', '--g2s_config', required=True, type=str, help='path to the config file')
    parser.add_argument('--grid_search', action='store_true', help='flag: grid search')
    args = vars(parser.parse_args())

    return args


def print_config(config):
    print('**************** MODEL CONFIGURATION ****************')
    for key in sorted(config.keys()):
        val = config[key]
        keystr = '{}'.format(key) + (' ' * (24 - len(key)))
        print('{} -->   {}'.format(keystr, val))
    print('**************** MODEL CONFIGURATION ****************')


def grid_search_main(config):
    grid_search_hyperparams = []
    log_path = config['out_dir']
    for k, v in config.items():
        if isinstance(v, list):
            grid_search_hyperparams.append(k)
            log_path += '_{}_{}'.format(k, v)

    logger = Logger(log_path, config=config, overwrite=True)

    best_config = None
    best_score = -1
    best_scores = None
    configs = grid(config)
    for cnf in configs:
        for k in grid_search_hyperparams:
            cnf['out_dir'] += '_{}_{}'.format(k, cnf[k])

        val_score, test_scores = main(cnf)
        if best_score < test_scores[cnf['early_stop_metric']]:
            best_score = test_scores[cnf['early_stop_metric']]
            best_scores = test_scores
            best_config = cnf
            print('Found a better configuration: {}'.format(best_scores))
            logger.write('Found a better configuration: {}'.format(best_scores))

    print('\nBest configuration:')
    logger.write('\nBest configuration:')
    for k in grid_search_hyperparams:
        print('{}: {}'.format(k, best_config[k]))
        logger.write('{}: {}'.format(k, best_config[k]))

    print('Best score: {}'.format(best_scores))
    logger.write('Best score: {}\n'.format(best_scores))
    logger.close()


if __name__ == '__main__':
    cfg = get_args()
    task_args = get_yaml_config(cfg['task_config'])
    g2s_args = get_yaml_config(cfg['g2s_config'])
    # load Graph2Seq template config
    g2s_template = get_basic_args(graph_construction_name=g2s_args['graph_construction_name'],
                              graph_embedding_name=g2s_args['graph_embedding_name'],
                              decoder_name=g2s_args['decoder_name'])
    update_values(to_args=g2s_template, from_args_list=[g2s_args, task_args])
    print_config(g2s_template)
    if cfg['grid_search']:
        grid_search_main(g2s_template)
    else:
        main(g2s_template)



import copy
import json

import torch
from pythonds.basic.stack import Stack
from stanfordcorenlp import StanfordCoreNLP

from graph4nlp.pytorch.data.data import GraphData, to_batch
from .base import StaticGraphConstructionBase
from .utils import CORENLP_TIMEOUT_SIGNATURE

"""TODO: some design choice:
            - replace constituent tag "." with "const_period"
            - unify whether use lower case in vocab and parsing
"""


class ConstituencyBasedGraphConstruction(StaticGraphConstructionBase):
    """
    Class for constituency graph construction.

    ...

    Attributes
    ----------
    embedding_styles : (dict)
        Specify embedding styles including ``single_token_item``, ``emb_strategy``, ``num_rnn_layers``, ``bert_model_name`` and ``bert_lower_case``.

    vocab: (set, optional)
        Vocabulary including all words appeared in graphs.

    Methods
    -------

    topology(raw_text_data, nlp_processor, merge_strategy=None, edge_strategy=None)
        Generate graph structure with nlp parser like ``CoreNLP`` etc.

    _construct_static_graph(parsed_object, sub_sentence_id, edge_strategy=None)
        Construct a single static graph from a single sentence, to be called by ``topology`` function.

    _graph_connect(nx_graph_list, merge_strategy=None)
        Construct a merged graph from a list of graphs, to be called by ``topology`` function.

    embedding(node_attributes, edge_attributes)
        Generate node/edge embeddings from node/edge attributes through an embedding layer.

    forward(raw_text_data, nlp_parser)
        Generate graph topology and embeddings.
    """

    def __init__(self, embedding_style, vocab, hidden_size, fix_word_emb=True, fix_bert_emb=True, word_dropout=None,
                 rnn_dropout=None, device=None):
        super(ConstituencyBasedGraphConstruction, self).__init__(word_vocab=vocab,
                                                                 embedding_styles=embedding_style,
                                                                 hidden_size=hidden_size,
                                                                 fix_word_emb=fix_word_emb,
                                                                 fix_bert_emb=fix_bert_emb,
                                                                 word_dropout=word_dropout,
                                                                 rnn_dropout=rnn_dropout,
                                                                 device=device)
        self.vocab = vocab
        assert (self.embedding_layer.device == device)
        self.device = self.embedding_layer.device

    @classmethod
    def parsing(cls, raw_text_data, nlp_processor, processor_args):
        '''
        Parameters
        ----------
        raw_text_data: str
        nlp_processor: StanfordCoreNLP
        split_hyphenated: bool
        normalize: bool
        '''
        output = nlp_processor.annotate(raw_text_data.strip(), properties=processor_args)
        if CORENLP_TIMEOUT_SIGNATURE in output:
            raise TimeoutError('CoreNLP timed out at input: \n{}\n This item will be skipped. '
                               'Please check the input or change the timeout threshold.'.format(raw_text_data))
        parsed_output = json.loads(output)['sentences']
        return parsed_output

    @classmethod
    def topology(cls,
                 raw_text_data,
                 nlp_processor,
                 processor_args,
                 merge_strategy=None,
                 edge_strategy=None,
                 verbase=True):
        """topology This function generate a graph strcuture from a raw text data.

        Parameters
        ----------
        raw_text_data : string
            A string to be used to construct a static graph, can be composed of multiple strings

        nlp_processor : object
            A parser used to parse sentence string to parsing trees like dependency parsing tree or constituency parsing tree

        merge_strategy : None or str, option=[None, "tailhead", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``: It will be the default option. We will do as ``"tailhead"``.
            ``"tailhead"``: Link the sub-graph  ``i``'s tail node with ``i+1``'s head node
            ``"user_define"``: We will give this option to the user. User can override this method to define your merge
                               strategy.

        edge_strategy: None or str, option=[None, "homogeneous", "heterogeneous", "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option. We will do as ``"homogeneous"``.
            ``"homogeneous"``: We will drop the edge type information.
                               If there is a linkage among node ``i`` and node ``j``, we will add an edge whose weight
                               is ``1.0``. Otherwise there is no edge.
            ``heterogeneous``: We will keep the edge type information.
                               An edge will have type information like ``n_subj``.
                               It is not implemented yet.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).
                         It is not implemented yet.

        Returns
        -------
        GraphData
            A customized graph data structure
        """
        output_graph_list = []
        parsed_output = cls.parsing(raw_text_data, nlp_processor, processor_args)
        for index in range(len(parsed_output)):
            output_graph_list.append(
                cls._construct_static_graph(parsed_output[index], index))
        ret_graph = cls._graph_connect(output_graph_list)
        if verbase:
            print('--------------------------------------')
            for _edge in ret_graph.get_all_edges():
                print(ret_graph.nodes[_edge[0]].attributes['token'],
                      "\t---\t", ret_graph.nodes[_edge[1]].attributes['token'])
            print('--------------------------------------')

        return ret_graph

    @classmethod
    def _construct_static_graph(cls,
                                parsed_object,
                                sub_sentence_id,
                                edge_strategy=None,
                                sequential_link=True,
                                bisequential_link=True,
                                top_down=False,
                                add_pos_node=False):
        """construct single syntactic graph

        Parameters
        ----------
        parsed_object :
            A parsed object of single sentence after external parser like ``CoreNLP``

        sub_sentence_id : int
            To specify which sentence the ``parsed_object`` is in the ``raw_text_data``

        bisequential_link : bool
            Add bi-directional edges between word nodes, do not add if ``False``

        top_down : bool
            Edge direction between nodes in constituency tree, if ``True``, add edges from top to down

        add_pos_node : bool
            Add part-of-speech nodes or not
        """
        parsed_sentence_data = parsed_object['parse']
        for punc in [u'(', u')']:
            parsed_sentence_data = parsed_sentence_data.replace(
                punc, ' ' + punc + ' ')
        parse_list = (parsed_sentence_data).split()

        res_graph = GraphData()
        pstack = Stack()
        idx = 0
        cnt_word_node = -1
        while idx < len(parse_list):
            if parse_list[idx] == '(':
                res_graph.add_nodes(1)
                res_graph.node_attributes[res_graph.get_node_num(
                ) - 1] = {'token': parse_list[idx + 1], 'type': 1, 'position_id': None, 'sentence_id': sub_sentence_id,
                          'tail': False, 'head': False}
                pstack.push(res_graph.get_node_num() - 1)
                if pstack.size() > 1:
                    node_2 = pstack.pop()
                    node_1 = pstack.pop()
                    if top_down:
                        res_graph.add_edge(node_1, node_2)
                    else:
                        res_graph.add_edge(node_2, node_1)
                    pstack.push(node_1)
                    pstack.push(node_2)
            elif parse_list[idx] == ')':
                pstack.pop()
            elif parse_list[idx + 1] == u')' and parse_list[idx] != u')':
                cnt_word_node += 1
                if add_pos_node:
                    res_graph.add_nodes(1)
                res_graph.node_attributes[res_graph.get_node_num(
                ) - 1] = {'token': parse_list[idx], 'type': 0, 'position_id': cnt_word_node,
                          'sentence_id': sub_sentence_id, 'tail': False, 'head': False}
                node_1 = pstack.pop()
                if node_1 != res_graph.get_node_num() - 1:
                    if top_down:
                        res_graph.add_edge(node_1, res_graph.get_node_num() - 1)
                    else:
                        res_graph.add_edge(res_graph.get_node_num() - 1, node_1)
                pstack.push(node_1)
            idx += 1

        if sequential_link:
            _len_single_graph = res_graph.get_node_num()
            _cnt_node = 0
            while (True):
                node_1_idx = -1
                node_2_idx = -1
                for idx, node_attr in res_graph.node_attributes.items():
                    if node_attr['position_id'] == _cnt_node:
                        node_1_idx = idx
                    elif node_attr['position_id'] == _cnt_node + 1:
                        node_2_idx = idx
                if node_1_idx != -1 and node_2_idx != -1:
                    res_graph.add_edge(node_1_idx, node_2_idx)
                    if bisequential_link:
                        res_graph.add_edge(node_2_idx, node_1_idx)
                if _cnt_node >= _len_single_graph:
                    break
                _cnt_node += 1

        max_pos = 0
        for n in res_graph.node_attributes.values():
            if n['type'] == 0 and n['position_id'] > max_pos:
                max_pos = n['position_id']

        min_pos = 99999
        for n in res_graph.node_attributes.values():
            if n['type'] == 0 and n['position_id'] < min_pos:
                min_pos = n['position_id']

        for n in res_graph.node_attributes.values():
            if n['type'] == 0 and n['position_id'] == max_pos:
                n['tail'] = True
            if n['type'] == 0 and n['position_id'] == min_pos:
                n['head'] = True
        # print(res_graph.edges)
        return res_graph

    @classmethod
    def _graph_connect(cls, graph_list, merge_strategy=None, bisequential_link=True, reformalize=True):
        """
        Parameters
        ----------
        graph_list : list
            A graph list to be merged

        bisequential_link : bool
            whether add bi-direnctional links between word nodes

        reformalize : bool
            If true, separate word nodes and non-terminal nodes in ``graph.node_attributes`` and put word nodes in the front position

        bisequential_link : bool
            whether add bi-direnctional links between word nodes

        reformalize : bool
            If true, separate word nodes and non-terminal nodes in ``graph.node_attributes`` and put word nodes in the front position

        Returns
        -------
        GraphData
            A customized graph data structure
        """
        _len_graph_ = len(graph_list)
        merged_graph = GraphData()
        for index in range(_len_graph_):
            len_merged_graph = merged_graph.get_node_num()

            graph_i_nodes_attributes = {}
            for dict_item in graph_list[index].node_attributes.items():
                graph_i_nodes_attributes[dict_item[0] +
                                         len_merged_graph] = dict_item[1]
            merged_graph.node_attributes.update(graph_i_nodes_attributes)

        _cnt_edge_num = 0
        for g_idx in range(_len_graph_):
            tmp_edges = graph_list[g_idx].get_all_edges()
            current_node_num = graph_list[g_idx].get_node_num()
            for _edge in tmp_edges:
                merged_graph.add_edge(
                    _edge[0] + _cnt_edge_num, _edge[1] + _cnt_edge_num)
            _cnt_edge_num += current_node_num

        for index in range(_len_graph_ - 1):
            # get head node
            head_node = -1
            for node_attr_item in merged_graph.node_attributes.items():
                if node_attr_item[1]['sentence_id'] == index + 1 and node_attr_item[1]['type'] == 0 and \
                        node_attr_item[1]['head'] == True:
                    head_node = node_attr_item[0]
            # get tail node
            tail_node = -1
            for node_attr_item in merged_graph.node_attributes.items():
                if node_attr_item[1]['sentence_id'] == index and node_attr_item[1]['type'] == 0 and node_attr_item[1][
                    'tail'] == True:
                    tail_node = node_attr_item[0]
            merged_graph.add_edge(tail_node, head_node)
            if bisequential_link:
                merged_graph.add_edge(head_node, tail_node)
        if reformalize:
            new_dict_for_word_nodes = copy.deepcopy(merged_graph.node_attributes)

            dict_for_word_nodes = {}
            dict_for_non_ternimal_nodes = {}
            for i in new_dict_for_word_nodes.items():
                if i[1]['type'] == 0:
                    dict_for_word_nodes[i[0]] = i[1]
                else:
                    dict_for_non_ternimal_nodes[i[0]] = i[1]
            merged_graph.node_attributes.clear()
            merged_graph.node_attributes.update({**dict_for_word_nodes, **dict_for_non_ternimal_nodes})

            node_id_map = {}
            cnt = 0
            for i in merged_graph.node_attributes.items():
                node_id_map[i[0]] = cnt
                cnt += 1

            for i in range(merged_graph.get_edge_num()):
                merged_graph._edge_indices.src[i] = node_id_map[merged_graph._edge_indices.src[i]]
                merged_graph._edge_indices.tgt[i] = node_id_map[merged_graph._edge_indices.tgt[i]]
            reformalize_graph_attributes = {}
            for i in merged_graph.node_attributes.items():
                reformalize_graph_attributes[node_id_map[i[0]]] = copy.deepcopy(i[1])

            merged_graph.node_attributes.clear()
            merged_graph.node_attributes.update(reformalize_graph_attributes)
        return merged_graph

    def forward(self, batch_graphdata: list):
        node_size = []
        num_nodes = []
        num_word_nodes = []  # number of nodes that are extracted from the raw text in each graph

        for g in batch_graphdata:
            g.node_features['token_id'] = g.node_features['token_id'].to(self.device)
            num_nodes.append(g.get_node_num())
            num_word_nodes.append(len([1 for i in range(len(g.node_attributes)) if g.node_attributes[i]['type'] == 0]))
            node_size.extend([1 for i in range(num_nodes[-1])])

        batch_gd = to_batch(batch_graphdata)
        node_size = torch.Tensor(node_size).to(self.device).int()
        num_nodes = torch.Tensor(num_nodes).to(self.device).int()
        num_word_nodes = torch.Tensor(num_word_nodes).to(self.device).int()
        node_emb = self.embedding_layer(batch_gd, node_size, num_nodes, num_word_items=num_word_nodes)
        batch_gd.node_features["node_feat"] = node_emb
        return batch_gd

    def embedding(self, node_attributes, edge_attributes):
        node_emb, edge_emb = self.embedding_layer(
            node_attributes, edge_attributes)
        return node_emb, edge_emb



import abc
import json
import os
import warnings
from collections import Counter
from multiprocessing import Pool

import numpy as np
import stanfordcorenlp
import torch.utils.data
from nltk.tokenize import word_tokenize
from sklearn import preprocessing
from copy import deepcopy

from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals_no_size
from ..data.data import GraphData
from ..modules.graph_construction.base import GraphConstructionBase
from ..modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from ..modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from ..modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
from ..modules.utils.tree_utils import Tree
from ..modules.utils.tree_utils import Vocab as VocabForTree
from ..modules.utils.vocab_utils import VocabModel, Vocab


class DataItem(object):
    def __init__(self, input_text, tokenizer):
        self.input_text = input_text
        self.tokenizer = tokenizer
        pass

    @abc.abstractmethod
    def extract(self):
        raise NotImplementedError


class Text2TextDataItem_seq2seq(DataItem):
    def __init__(self, input_text, output_text, tokenizer, share_vocab=True):
        super(Text2TextDataItem_seq2seq, self).__init__(input_text, tokenizer)
        self.output_text = output_text
        self.share_vocab = share_vocab

    def extract(self, lower_case=True):
        """
        Returns
        -------
        Input tokens and output tokens
        """

        if lower_case:
            self.input_text = self.input_text.lower()

        if self.tokenizer is None:
            input_tokens = self.input_text.strip().split(' ')
        else:
            input_tokens = self.tokenizer(self.input_text)


        if lower_case:
            self.output_text = self.output_text.lower()

        if self.tokenizer is None:
            output_tokens = self.output_text.strip().split(' ')
        else:
            if '<t>' in self.output_text:
                output_text = self.output_text.replace('<t>','').replace('</t>','')
            output_tokens = self.tokenizer(output_text)
            output_tokens = output_tokens + ['<t>','</t>']

        if self.share_vocab:
            return input_tokens + output_tokens
        else:
            return input_tokens, output_tokens


class Text2TextDataItem(DataItem):
    def __init__(self, input_text, output_text, tokenizer, share_vocab=True):
        super(Text2TextDataItem, self).__init__(input_text, tokenizer)
        self.output_text = output_text
        self.share_vocab = share_vocab

    def extract(self):
        """
        Returns
        -------
        Input tokens and output tokens
        """
        g: GraphData = self.graph

        input_tokens = []
        for i in range(g.get_node_num()):
            if self.tokenizer is None:
                tokenized_token = g.node_attributes[i]['token'].strip().split(' ')
            else:
                tokenized_token = self.tokenizer(g.node_attributes[i]['token'])

            input_tokens.extend(tokenized_token)

        if self.tokenizer is None:
            output_tokens = self.output_text.strip().split(' ')
        else:
            output_tokens = self.tokenizer(self.output_text)

        if self.share_vocab:
            return input_tokens + output_tokens
        else:
            return input_tokens, output_tokens


class Text2TreeDataItem(DataItem):
    def __init__(self, input_text, output_text, output_tree, tokenizer, share_vocab=True):
        super(Text2TreeDataItem, self).__init__(input_text, tokenizer)
        self.output_text = output_text
        self.share_vocab = share_vocab
        self.output_tree = output_tree
    def extract(self, use_tok=True):
        """
        Returns
        -------
        Input tokens and output tokens
        """
        g: GraphData = self.graph

        input_tokens = []
        for i in range(g.get_node_num()):
            if self.tokenizer is None or use_tok == False:
                tokenized_token = g.node_attributes[i]['token'].strip().split(' ')
            else:
                tokenized_token = self.tokenizer(g.node_attributes[i]['token'])

            input_tokens.extend(tokenized_token)

        if self.tokenizer is None or use_tok == False:
            output_tokens = self.output_text.strip().split(' ')
        else:
            output_tokens = self.tokenizer(self.output_text)

        return input_tokens, output_tokens


class Text2LabelDataItem(DataItem):
    def __init__(self, input_text, output_label, tokenizer):
        super(Text2LabelDataItem, self).__init__(input_text, tokenizer)
        self.output_label = output_label

    def extract(self):
        """
        Returns
        -------
        Input tokens and output tokens
        """
        g: GraphData = self.graph

        input_tokens = []
        for i in range(g.get_node_num()):
            if self.tokenizer is None:
                tokenized_token = g.node_attributes[i]['token'].strip().split()
            else:
                tokenized_token = self.tokenizer(g.node_attributes[i]['token'])

            input_tokens.extend(tokenized_token)

        return input_tokens


class DoubleText2TextDataItem(DataItem):
    def __init__(self, input_text, input_text2, output_text, tokenizer, share_vocab=True):
        super(DoubleText2TextDataItem, self).__init__(input_text, tokenizer)
        self.input_text2 = input_text2
        self.output_text = output_text
        self.share_vocab = share_vocab

    def extract(self):
        """
        Returns
        -------
        Input tokens and output tokens
        """
        g: GraphData = self.graph

        input_tokens = []
        for i in range(g.get_node_num()):
            if self.tokenizer is None:
                tokenized_token = g.node_attributes[i]['token'].strip().split()
            else:
                tokenized_token = self.tokenizer(g.node_attributes[i]['token'])

            input_tokens.extend(tokenized_token)

        if self.tokenizer is None:
            input_tokens.extend(self.input_text2.strip().split())
            output_tokens = self.output_text.strip().split()
        else:
            input_tokens.extend(self.tokenizer(self.input_text2))
            output_tokens = self.tokenizer(self.output_text)

        if self.share_vocab:
            return input_tokens + output_tokens
        else:
            return input_tokens, output_tokens


class SequenceLabelingDataItem(DataItem):
    def __init__(self, input_text, output_tags, tokenizer):
        super(SequenceLabelingDataItem, self).__init__(input_text, tokenizer)
        self.output_tag = output_tags

    def extract(self):
        """
        Returns
        -------
        Input tokens and output tags
        """
        g: GraphData = self.graph

        input_tokens = []
        for i in range(g.get_node_num()):
            if self.tokenizer is None:
                tokenized_token = self.output_text.strip().split(' ')
            else:
                tokenized_token = self.tokenizer(g.node_attributes[i]['token'])

            input_tokens.extend(tokenized_token)

        return input_tokens


class Dataset(torch.utils.data.Dataset):
    """
    Base class for datasets.

    The dataset is organized in a two-layer index style. Direct access to the dataset object, e.g. Dataset[1], will first
    be converted to the access to the internal index list, which is then passed to access the actual data. This design
    is for the ease of sampling.

    Examples
    --------
    Suppose we have a Dataset containing 5 data items ['a', 'b', 'c', 'd', 'e']. The indices of the 5 elements in the
    list are correspondingly [0, 1, 2, 3, 4]. Suppose the dataset is shuffled, which shuffles the internal index list, the
    consequent indices becomes [2, 3, 1, 4, 5]. Then an access to the dataset `Dataset[2]` will first access the indices[2]
    which is 1, and then use the received index to access the actual dataset, which will return the actual data item 'b'.
    Now to the user the 3rd ([2]) element in the dataset got shuffled and is not 'c'.

    Parameters
    ----------
    root: str
        The root directory path where the dataset is stored.
    """

    @property
    def raw_file_names(self) -> dict:
        raise NotImplementedError

    @property
    def processed_file_names(self) -> dict:
        raise NotImplementedError

    @abc.abstractmethod
    def download(self):
        """Download the raw data from the Internet."""
        raise NotImplementedError

    @abc.abstractmethod
    def vectorization(self, data_items):
        """Convert tokens to indices which can be processed by downstream models."""
        raise NotImplementedError

    @abc.abstractmethod
    def parse_file(self, file_path):
        """To be implemented in task-specific dataset base class."""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def collate_fn(data_list):
        """Takes a list of data and convert it to a batch of data."""
        raise NotImplementedError

    def __init__(self,
                 root,
                 topology_builder,
                 topology_subdir,
                 tokenizer=word_tokenize,
                 lower_case=True,
                 pretrained_word_emb_file=None,
                 use_val_for_vocab=False,
                 seed=1234,
                 device='cpu',
                 thread_number=4,
                 port=9000,
                 timeout=15000,
                 **kwargs):
        """

        Parameters
        ----------
        root: str
            The path of the data root.
        topology_builder: GraphConstructionBase
            The initial graph topology builder.
        topology_subdir: str
            The name of the data folder.
        tokenizer: function, default=nltk.word_tokenize
            The word tokenizer.
        lower_case: bool, default=True
            Whether to use lower-case option.
        pretrained_word_emb_file: str, default=None
            The path of the pre-trained word embedding file.
        use_val_for_vocab: bool, default=False
            Whether to add val split in the final split.
        seed: int, default=1234
            The seed for random function.
        device: str, default='cpu'
            The device of the current environment.
        thread_number: int, default=4
            The thread number for building initial graph. For most case, it may be the number of your CPU cores.
        port: int, default=9000
            The port for stanfordcorenlp.
        timeout: int, default=15000
            The timeout for stanfordcorenlp.
        kwargs
        """
        super(Dataset, self).__init__()

        self.root = root  # The root directory where the dataset is located.
        self.seed = seed

        # stanfordcorenlp hyper-parameter
        self.thread_number = thread_number
        self.port = port
        self.timeout = timeout

        # Processing-specific attributes
        self.tokenizer = tokenizer
        self.lower_case = lower_case
        self.pretrained_word_emb_file = pretrained_word_emb_file
        self.topology_builder = topology_builder
        self.topology_subdir = topology_subdir
        self.use_val_for_vocab = use_val_for_vocab
        self.device = device
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.__indices__ = None

        if 'download' in self.__class__.__dict__.keys():
            self._download()

        self._process()

        # After initialization, load the preprocessed files.
        data = torch.load(self.processed_file_paths['data'])
        self.train = data['train']
        self.test = data['test']
        if 'val' in data.keys():
            self.val = data['val']

        if 'KG_graph' in self.processed_file_paths.keys():
            self.KG_graph = torch.load(self.processed_file_paths['KG_graph'])

        self.build_vocab()

    @property
    def raw_dir(self) -> str:
        """The directory where the raw data is stored."""
        return os.path.join(self.root, 'raw')

    @property
    def processed_dir(self) -> str:
        return os.path.join(self.root, 'processed', self.topology_subdir)

    @property
    def raw_file_paths(self) -> dict:
        """The paths to raw files."""
        return {key: os.path.join(self.raw_dir, name) for key, name in self.raw_file_names.items()}

    @property
    def processed_file_paths(self) -> dict:
        return {name: os.path.join(self.processed_dir, processed_file_name) for name, processed_file_name in
                self.processed_file_names.items()}

    def _download(self):
        if all([os.path.exists(raw_path) for raw_path in self.raw_file_paths.values()]):
            return

        os.makedirs(self.raw_dir, exist_ok=True)
        self.download()

    def read_raw_data(self):
        """
        Read raw data from the disk and put them in a dictionary (`self.data`).
        The raw data file should be organized as the format defined in `self.parse_file()` method.

        This function calls `self.parse_file()` repeatedly and pass the file paths in `self.raw_file_names` once at a time.

        This function builds `self.data` which is a dict of {int (index): DataItem}, where the id represents the
        index of the DataItem w.r.t. the whole dataset.

        This function also builds the `self.split_ids` dictionary whose keys correspond to those of self.raw_file_names
        defined by the user, indicating the indices of each subset (e.g. train, val and test).

        """
        self.train = self.parse_file(self.raw_file_paths['train'])
        self.test = self.parse_file(self.raw_file_paths['test'])
        if 'val' in self.raw_file_paths.keys():
            self.val = self.parse_file(self.raw_file_paths['val'])
        elif 'val_split_ratio' in self.__dict__:
            if self.val_split_ratio > 0:
                new_train_length = int((1 - self.val_split_ratio) * len(self.train))
                import random
                random.seed(self.seed)
                old_train_set = self.train
                random.shuffle(old_train_set)
                self.val = old_train_set[new_train_length:]
                self.train = old_train_set[:new_train_length]

    @staticmethod
    def _build_topology_process(data_items, topology_builder,
                                graph_type, dynamic_graph_type, dynamic_init_topology_builder,
                                merge_strategy, edge_strategy, dynamic_init_topology_aux_args,
                                lower_case, tokenizer, port, timeout):

        ret = []
        if graph_type == 'static':
            print('Connecting to stanfordcorenlp server...')
            processor = stanfordcorenlp.StanfordCoreNLP('http://localhost', port=port, timeout=timeout)

            if topology_builder == IEBasedGraphConstruction:
                props_coref = {
                    'annotators': 'tokenize, ssplit, pos, lemma, ner, parse, coref',
                    "tokenize.options":
                        "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
                props_openie = {
                    'annotators': 'tokenize, ssplit, pos, ner, parse, openie',
                    "tokenize.options":
                        "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json',
                    "openie.triple.strict": "true"
                }
                processor_args = [props_coref, props_openie]
            elif topology_builder == DependencyBasedGraphConstruction:
                processor_args = {
                    'annotators': 'ssplit,tokenize,depparse',
                    "tokenize.options":
                        "splitHyphenated=false,normalizeParentheses=false,normalizeOtherBrackets=false",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
            elif topology_builder == ConstituencyBasedGraphConstruction:
                processor_args = {
                    'annotators': "tokenize,ssplit,pos,parse",
                    "tokenize.options":
                        "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
            else:
                raise NotImplementedError
            print('CoreNLP server connected.')
            pop_idxs = []
            for cnt, item in enumerate(data_items):
                if cnt % 1000 == 0:
                    print("Port {}, processing: {} / {}".format(port, cnt, len(data_items)))
                try:
                    graph = topology_builder.topology(raw_text_data=item.input_text,
                                                      nlp_processor=processor,
                                                      processor_args=processor_args,
                                                      merge_strategy=merge_strategy,
                                                      edge_strategy=edge_strategy,
                                                      verbase=False)
                    item.graph = graph
                except Exception as msg:
                    pop_idxs.append(cnt)
                    item.graph = None
                    warnings.warn(RuntimeWarning(msg))
                ret.append(item)
            ret = [x for idx, x in enumerate(ret) if idx not in pop_idxs]
        elif graph_type == 'dynamic':
            if dynamic_graph_type == 'node_emb':
                for item in data_items:
                    graph = topology_builder.init_topology(item.input_text,
                                                           lower_case=lower_case,
                                                           tokenizer=tokenizer)
                    ret.append(graph)
            elif dynamic_graph_type == 'node_emb_refined':
                if dynamic_init_topology_builder in (
                IEBasedGraphConstruction, DependencyBasedGraphConstruction, ConstituencyBasedGraphConstruction):
                    print('Connecting to stanfordcorenlp server...')
                    processor = stanfordcorenlp.StanfordCoreNLP('http://localhost', port=port, timeout=timeout)

                    if dynamic_init_topology_builder == IEBasedGraphConstruction:
                        props_coref = {
                            'annotators': 'tokenize, ssplit, pos, lemma, ner, parse, coref',
                            "tokenize.options":
                                "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                        props_openie = {
                            'annotators': 'tokenize, ssplit, pos, ner, parse, openie',
                            "tokenize.options":
                                "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json',
                            "openie.triple.strict": "true"
                        }
                        processor_args = [props_coref, props_openie]
                    elif dynamic_init_topology_builder == DependencyBasedGraphConstruction:
                        processor_args = {
                            'annotators': 'ssplit,tokenize,depparse',
                            "tokenize.options":
                                "splitHyphenated=false,normalizeParentheses=false,normalizeOtherBrackets=false",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                    elif dynamic_init_topology_builder == ConstituencyBasedGraphConstruction:
                        processor_args = {
                            'annotators': "tokenize,ssplit,pos,parse",
                            "tokenize.options":
                                "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                    else:
                        raise NotImplementedError
                    print('CoreNLP server connected.')
                else:
                    processor = None
                    processor_args = None
                pop_idxs = []
                for idx, item in enumerate(data_items):
                    try:
                        graph = topology_builder.init_topology(item.input_text,
                                                               dynamic_init_topology_builder=dynamic_init_topology_builder,
                                                               lower_case=lower_case,
                                                               tokenizer=tokenizer,
                                                               nlp_processor=processor,
                                                               processor_args=processor_args,
                                                               merge_strategy=merge_strategy,
                                                               edge_strategy=edge_strategy,
                                                               verbase=False,
                                                               dynamic_init_topology_aux_args=dynamic_init_topology_aux_args)

                        item.graph = graph
                    except Exception as msg:
                        pop_idxs.append(idx)
                        item.graph = None
                        warnings.warn(RuntimeWarning(msg))
                    ret.append(item)
                ret = [x for idx, x in enumerate(ret) if idx not in pop_idxs]
            else:
                raise RuntimeError('Unknown dynamic_graph_type: {}'.format(dynamic_graph_type))

        else:
            raise NotImplementedError('Currently only static and dynamic are supported!')

        return ret

    def build_topology(self, data_items):
        """
        Build graph topology for each item in the dataset. The generated graph is bound to the `graph` attribute of the
        DataItem.
        """
        total = len(data_items)
        thread_number = min(total, self.thread_number)
        pool = Pool(thread_number)
        res_l = []
        for i in range(thread_number):
            start_index = total * i // thread_number
            end_index = total * (i + 1) // thread_number

            """
            data_items, topology_builder,
                                graph_type, dynamic_graph_type, dynamic_init_topology_builder,
                                merge_strategy, edge_strategy, dynamic_init_topology_aux_args,
                                lower_case, tokenizer, port, timeout
            """
            r = pool.apply_async(self._build_topology_process,
                                 args=(data_items[start_index:end_index], self.topology_builder, self.graph_type,
                                       self.dynamic_graph_type, self.dynamic_init_topology_builder,
                                       self.merge_strategy, self.edge_strategy, self.dynamic_init_topology_aux_args,
                                       self.lower_case, self.tokenizer, self.port, self.timeout))
            res_l.append(r)
        pool.close()
        pool.join()

        data_items = []
        for i in range(thread_number):
            res = res_l[i].get()
            for data in res:
                data.graph = data.graph.to(self.device)
                data_items.append(data)

        return data_items

    def build_vocab(self):
        """
        Build the vocabulary. If `self.use_val_for_vocab` is `True`, use both training set and validation set for building
        the vocabulary. Otherwise only the training set is used.

        """
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = self.val + data_for_vocab

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=self.max_word_vocab_size,
                                       min_word_vocab_freq=self.min_word_vocab_size,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=self.word_emb_size)
        self.vocab_model = vocab_model

        return self.vocab_model

    def _process(self):
        if all([os.path.exists(processed_path) for processed_path in self.processed_file_paths.values()]):
            if 'val_split_ratio' in self.__dict__:
                UserWarning(
                    "Loading existing processed files on disk. Your `val_split_ratio` might not work since the data have"
                    "already been split.")
            return

        os.makedirs(self.processed_dir, exist_ok=True)

        self.read_raw_data()

        self.train = self.build_topology(self.train)
        self.test = self.build_topology(self.test)
        if 'val' in self.__dict__:
            self.val = self.build_topology(self.val)

        self.build_vocab()

        self.vectorization(self.train)
        self.vectorization(self.test)
        if 'val' in self.__dict__:
            self.vectorization(self.val)

        data_to_save = {'train': self.train, 'test': self.test}
        if 'val' in self.__dict__:
            data_to_save['val'] = self.val
        torch.save(data_to_save, self.processed_file_paths['data'])


class Text2TextDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir, share_vocab=True, **kwargs):
        self.data_item_type = Text2TextDataItem
        self.share_vocab = share_vocab
        super(Text2TextDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.

        For Text2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).

        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )

        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"

        Parameters
        ----------
        file_path: str
            The path of the input file.

        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                input, output = line.split('\t')
                data_item = Text2TextDataItem(input_text=input, output_text=output, tokenizer=self.tokenizer,
                                              share_vocab=self.share_vocab)
                data.append(data_item)
        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=None,
                                       min_word_vocab_freq=1,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=300,
                                       share_vocab=self.share_vocab)
        self.vocab_model = vocab_model

        return self.vocab_model

    def vectorization(self, data_items):
        if self.topology_builder == IEBasedGraphConstruction:
            use_ie = True
        else:
            use_ie = False
        for item in data_items:
            graph: GraphData = item.graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.vocab_model.in_word_vocab.getIndex(node_token, use_ie)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            if self.topology_builder == IEBasedGraphConstruction:
                for i in range(len(token_matrix)):
                    token_matrix[i] = np.array(token_matrix[i][0])
                token_matrix = pad_2d_vals_no_size(token_matrix)
                token_matrix = torch.tensor(token_matrix, dtype=torch.long)
                graph.node_features['token_id'] = token_matrix
                pass
            else:
                token_matrix = torch.tensor(token_matrix, dtype=torch.long)
                graph.node_features['token_id'] = token_matrix

            if use_ie and 'token' in graph.edge_attributes[0].keys():
                edge_token_matrix = []
                for edge_idx in range(graph.get_edge_num()):
                    edge_token = graph.edge_attributes[edge_idx]['token']
                    edge_token_id = self.vocab_model.in_word_vocab.getIndex(edge_token, use_ie)
                    graph.edge_attributes[edge_idx]['token_id'] = edge_token_id
                    edge_token_matrix.append([edge_token_id])
                if self.topology_builder == IEBasedGraphConstruction:
                    for i in range(len(edge_token_matrix)):
                        edge_token_matrix[i] = np.array(edge_token_matrix[i][0])
                    edge_token_matrix = pad_2d_vals_no_size(edge_token_matrix)
                    edge_token_matrix = torch.tensor(edge_token_matrix, dtype=torch.long)
                    graph.edge_features['token_id'] = edge_token_matrix

            tgt = item.output_text
            tgt_token_id = self.vocab_model.out_word_vocab.to_index_sequence(tgt)
            tgt_token_id.append(self.vocab_model.out_word_vocab.EOS)
            tgt_token_id = np.array(tgt_token_id)
            item.output_np = tgt_token_id

    @staticmethod
    def collate_fn(data_list: [Text2TextDataItem]):
        graph_data = [deepcopy(item.graph) for item in data_list]

        output_numpy = [deepcopy(item.output_np) for item in data_list]
        output_str = [deepcopy(item.output_text.lower().strip()) for item in data_list]
        output_pad = pad_2d_vals_no_size(output_numpy)

        from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals
        max_num_tokens_a_node = max([x.graph.node_features['token_id'].size()[1] for x in data_list])
        if max_num_tokens_a_node>1:
            for x in data_list:
                x.graph.node_features['token_id'] = torch.from_numpy(
                    pad_2d_vals(x.graph.node_features['token_id'].cpu().numpy(),
                                x.graph.node_features['token_id'].size()[0],
                                max_num_tokens_a_node)).long()

        tgt_seq = torch.from_numpy(output_pad).long()
        # return [graph_data, tgt_seq, output_str]
        return {
            "graph_data": graph_data,
            "tgt_seq": tgt_seq,
            "output_str": output_str
        }


class TextToTreeDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir, min_freq, share_vocab=True, **kwargs):
        self.data_item_type = Text2TreeDataItem
        self.share_vocab = share_vocab
        self.min_freq = min_freq
        super(TextToTreeDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.

        For Text2TreeDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).

        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )

        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"

        Parameters
        ----------
        file_path: str
            The path of the input file.

        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                input, output = line.split('\t')
                data_item = Text2TreeDataItem(input_text=input, output_text=output, output_tree=None,
                                              tokenizer=self.tokenizer,
                                              share_vocab=self.share_vocab)
                data.append(data_item)
        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        src_vocab_model = VocabForTree(lower_case=self.lower_case,
                                       pretrained_embedding_fn=self.pretrained_word_emb_file,
                                       embedding_dims=self.enc_emb_size)
        tgt_vocab_model = VocabForTree(lower_case=self.lower_case,
                                       pretrained_embedding_fn=self.pretrained_word_emb_file,
                                       embedding_dims=self.dec_emb_size)

        if self.share_vocab:
            all_words = Counter()
        else:
            all_words = [Counter(), Counter()]

        for instance in data_for_vocab:
            extracted_tokens = instance.extract(use_tok=False)
            if self.share_vocab:
                all_words.update(extracted_tokens[0])
                for w_i in extracted_tokens[1]:
                    all_words[w_i] = 100000
            else:
                all_words[0].update(extracted_tokens[0])
                all_words[1].update(extracted_tokens[1])

        if self.share_vocab:
            print("min_freq_for_word", self.min_freq)
            src_vocab_model.init_from_list(list(all_words.items()), min_freq=self.min_freq, max_vocab_size=100000)
            tgt_vocab_model = src_vocab_model
        else:
            src_vocab_model.init_from_list(list(all_words[0].items()), min_freq=3, max_vocab_size=100000)
            tgt_vocab_model.init_from_list(list(all_words[1].items()), min_freq=1, max_vocab_size=100000)

        self.src_vocab_model = src_vocab_model
        self.tgt_vocab_model = tgt_vocab_model
        return self.src_vocab_model

    def vectorization(self, data_items):
        for item in data_items:
            graph: GraphData = item.graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.src_vocab_model.get_symbol_idx(node_token)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            token_matrix = torch.tensor(token_matrix, dtype=torch.long)
            graph.node_features['token_id'] = token_matrix

            tgt = item.output_text
            tgt_list = self.tgt_vocab_model.get_symbol_idx_for_list(tgt.split())
            output_tree = Tree.convert_to_tree(tgt_list, 0, len(tgt_list), self.tgt_vocab_model)
            item.output_tree = output_tree

    @staticmethod
    def collate_fn(data_list: [Text2TreeDataItem]):
        graph_data = [deepcopy(item.graph) for item in data_list]
        output_tree_list = [deepcopy(item.output_tree) for item in data_list]
        return [graph_data, output_tree_list]


class KGDataItem(DataItem):
    def __init__(self, e1, rel, e2, rel_eval, e2_multi, e1_multi, share_vocab=True, split_token=' '):
        super(KGDataItem, self).__init__(input_text=None, tokenizer=None)
        self.e1 = e1
        self.rel = rel
        self.e2 = e2
        self.rel_eval = rel_eval
        self.e2_multi = e2_multi
        self.e1_multi = e1_multi
        self.share_vocab = share_vocab
        self.split_token = split_token

    def extract(self):
        """
        Returns
        -------
        Input tokens and output tokens
        """
        # g: GraphData = self.graph

        input_tokens = []
        if self.tokenizer is None:
            e1_tokens = self.e1.strip().split(' ')
        else:
            e1_tokens = self.tokenizer(self.e1)

        if self.tokenizer is None:
            e2_tokens = self.e2.strip().split(' ')
        else:
            e2_tokens = self.tokenizer(self.e2)

        if self.tokenizer is None:
            e2_multi_tokens = self.e2_multi.strip().split(' ')
        else:
            e2_multi_tokens = self.tokenizer(self.e2_multi)

        if self.tokenizer is None:
            e1_multi_tokens = self.e1_multi.strip().split(' ')
        else:
            e1_multi_tokens = self.tokenizer(self.e1_multi)

        if self.tokenizer is None:
            rel_tokens = self.rel.strip().split(self.split_token)
        else:
            rel_tokens = self.tokenizer(self.rel)

        if self.tokenizer is None:
            rel_eval_tokens = self.rel_eval.strip().split(self.split_token)
        else:
            rel_eval_tokens = self.tokenizer(self.rel_eval)

        if self.share_vocab:
            return e1_tokens + e2_tokens + e1_multi_tokens + e2_multi_tokens + rel_tokens + rel_eval_tokens
        else:
            return e1_tokens + e2_tokens + e1_multi_tokens + e2_multi_tokens, rel_tokens + rel_eval_tokens


class KGCompletionDataset(Dataset):
    def __init__(self, root_dir, topology_builder=None, topology_subdir=None, share_vocab=True,
                 edge_strategy=None, **kwargs):
        self.data_item_type = KGDataItem
        self.share_vocab = share_vocab  # share vocab between entity and relation
        self.edge_strategy = edge_strategy
        super(KGCompletionDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.

        For Text2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).

        Examples
        --------
        {"e1": "person84", "e2": "person85", "rel": "term21", "rel_eval": "term21_reverse",
        "e2_multi1": "person85",
        "e2_multi2": "person74 person84 person55 person96 person66 person57"}

        {"e1": "person20", "e2": "person90", "rel": "term11", "rel_eval": "term11_reverse",
        "e2_multi1": "person29 person82 person85 person77 person73 person63 person34 person86 person4
        person83 person46 person16 person48 person17 person59 person80 person50 person90",
        "e2_multi2": "person29 person82 person2 person20 person83 person46 person80"}

        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"

        Parameters
        ----------
        file_path: str
            The path of the input file.

        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                line_dict = json.loads(line)
                data_item = KGDataItem(e1=line_dict['e1'],
                                       e2=line_dict['e2'],
                                       rel=line_dict['rel'],
                                       rel_eval=line_dict['rel_eval'],
                                       e2_multi=line_dict['e2_multi1'],
                                       e1_multi=line_dict['e2_multi2'],
                                       share_vocab=self.share_vocab,
                                       split_token=self.split_token)
                data.append(data_item)

                if 'train' in file_path:
                    for e2 in line_dict['e2_multi1'].split(' '):
                        triple = [line_dict['e1'], line_dict['rel'], e2]
                        self.build_parsed_results(triple)

                    for e1 in line_dict['e2_multi2'].split(' '):
                        triple = [line_dict['e2'], line_dict['rel_eval'], e1]
                        self.build_parsed_results(triple)

        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(
            saved_vocab_file=os.path.join(self.processed_dir, self.processed_file_names['vocab']),
            data_set=data_for_vocab,
            tokenizer=self.tokenizer,
            lower_case=self.lower_case,
            max_word_vocab_size=None,
            min_word_vocab_freq=1,
            pretrained_word_emb_file=self.pretrained_word_emb_file,
            word_emb_size=300)
        self.vocab_model = vocab_model

        return self.vocab_model

    def build_parsed_results(self, triple):
        if self.edge_strategy is None:
            if triple[0] not in self.graph_nodes:
                self.graph_nodes.append(triple[0])

            if triple[2] not in self.graph_nodes:
                self.graph_nodes.append(triple[2])

            if triple[1] not in self.graph_edges:
                self.graph_edges.append(triple[1])

            triple_info = {'edge_tokens': triple[1],
                           'src': {
                               'tokens': triple[0],
                               'id': self.graph_nodes.index(triple[0])
                           },
                           'tgt': {
                               'tokens': triple[2],
                               'id': self.graph_nodes.index(triple[2])
                           }}
            if triple_info not in self.parsed_results['graph_content']:
                self.parsed_results['graph_content'].append(triple_info)
        elif self.edge_strategy == "as_node":
            if triple[0] not in self.graph_nodes:
                self.graph_nodes.append(triple[0])

            if triple[1] not in self.graph_nodes:
                self.graph_nodes.append(triple[1])

            if triple[2] not in self.graph_nodes:
                self.graph_nodes.append(triple[2])

            triple_info_0_1 = {'edge_tokens': [],
                               'src': {
                                   'tokens': triple[0],
                                   'id': self.graph_nodes.index(triple[0]),
                                   'type': 'ent_node'
                               },
                               'tgt': {
                                   'tokens': triple[1],
                                   'id': self.graph_nodes.index(triple[1]),
                                   'type': 'edge_node'
                               }}

            triple_info_1_2 = {'edge_tokens': [],
                               'src': {
                                   'tokens': triple[1],
                                   'id': self.graph_nodes.index(triple[1]),
                                   'type': 'edge_node'
                               },
                               'tgt': {
                                   'tokens': triple[2],
                                   'id': self.graph_nodes.index(triple[2]),
                                   'type': 'ent_node'
                               }}

            if triple_info_0_1 not in self.parsed_results['graph_content']:
                self.parsed_results['graph_content'].append(triple_info_0_1)
            if triple_info_1_2 not in self.parsed_results['graph_content']:
                self.parsed_results['graph_content'].append(triple_info_1_2)

        return

    def build_topology(self):
        # self.data: [KGDataItem] = self.build_dataitem(self.raw_file_paths)
        self.KG_graph = GraphData()
        self.parsed_results['node_num'] = len(self.graph_nodes)
        self.parsed_results['graph_nodes'] = self.graph_nodes
        self.KG_graph = IEBasedGraphConstruction._construct_static_graph(self.parsed_results,
                                                                         edge_strategy=self.edge_strategy)
        self.KG_graph.graph_attributes['num_entities'] = len(self.graph_nodes)
        self.KG_graph.graph_attributes['num_relations'] = len(self.graph_edges)
        self.KG_graph.graph_attributes['graph_nodes'] = self.graph_nodes
        self.KG_graph.graph_attributes['graph_edges'] = self.graph_edges

    def _process(self):
        if all([os.path.exists(processed_path) for processed_path in self.processed_file_paths.values()]):
            if 'val_split_ratio' in self.__dict__:
                UserWarning(
                    "Loading existing processed files on disk. Your `val_split_ratio` might not work since the data have"
                    "already been split.")
            return

        os.makedirs(self.processed_dir, exist_ok=True)

        self.graph_nodes = []
        self.graph_edges = []
        """
        `self.parsed_results` is an intermediate dict that contains all the information of the KG graph.
        `self.parsed_results['graph_content']` is a list of dict.

        Each dict in `self.parsed_results['graph_content']` contains information about a triple
        (src_ent, rel, tgt_ent).

        `self.parsed_results['graph_nodes']` contains all nodes in the KG graph.
        `self.parsed_results['node_num']` is the number of nodes in the KG graph.
        """
        self.parsed_results = {}
        self.parsed_results['graph_content'] = []

        self.read_raw_data()
        self.build_topology()
        self.build_vocab()

        self.vec_graph = False
        self.vectorization(self.train)
        self.vectorization(self.test)
        if 'val' in self.__dict__:
            self.vectorization(self.val)

        data_to_save = {'train': self.train, 'test': self.test}
        if 'val' in self.__dict__:
            data_to_save['val'] = self.val
        torch.save(data_to_save, self.processed_file_paths['data'])

    def vectorization(self, data_items):
        if self.vec_graph == False:
            graph: GraphData = self.KG_graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.vocab_model.in_word_vocab.getIndex(node_token)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            token_matrix = torch.tensor(token_matrix, dtype=torch.long)
            graph.node_features['token_id'] = token_matrix
            self.vec_graph = True
            torch.save(self.KG_graph, os.path.join(self.processed_dir, self.processed_file_names['KG_graph']))

        for item in data_items:
            e1 = item.e1
            item.e1_tensor = torch.tensor(self.graph_nodes.index(e1), dtype=torch.long)

            e2 = item.e2
            item.e2_tensor = torch.tensor(self.graph_nodes.index(e2), dtype=torch.long)

            rel = item.rel
            if self.edge_strategy == "as_node":
                item.rel_tensor = torch.tensor(self.graph_nodes.index(rel), dtype=torch.long)
            else:
                item.rel_tensor = torch.tensor(self.graph_edges.index(rel), dtype=torch.long)

            rel_eval = item.rel_eval
            item.rel_eval_tensor = torch.tensor(self.graph_edges.index(rel_eval), dtype=torch.long)

            e2_multi = item.e2_multi
            item.e2_multi_tensor = torch.zeros(1, len(self.graph_nodes)). \
                scatter_(1,
                         torch.tensor([self.graph_nodes.index(i) for i in e2_multi.split()], dtype=torch.long).view(1,
                                                                                                                    -1),
                         torch.ones(1, len(e2_multi.split()))).squeeze()

            item.e2_multi_tensor_idx = torch.tensor([self.graph_nodes.index(i) for i in e2_multi.split()],
                                                    dtype=torch.long)

            e1_multi = item.e1_multi
            item.e1_multi_tensor = torch.zeros(1, len(self.graph_nodes)). \
                scatter_(1,
                         torch.tensor([self.graph_nodes.index(i) for i in e1_multi.split()], dtype=torch.long).view(1,
                                                                                                                    -1),
                         torch.ones(1, len(e1_multi.split()))).squeeze()
            item.e1_multi_tensor_idx = torch.tensor([self.graph_nodes.index(i) for i in e1_multi.split()],
                                                    dtype=torch.long)

    @staticmethod
    def collate_fn(data_list: [KGDataItem]):
        data_list = deepcopy(data_list)
        # graph_data = [item.graph for item in data_list]
        e1 = torch.tensor([item.e1_tensor for item in data_list])
        e2 = torch.tensor([item.e2_tensor for item in data_list])
        rel = torch.tensor([item.rel_tensor for item in data_list])
        rel_eval = torch.tensor([item.rel_eval_tensor for item in data_list])

        # do padding here
        e1_multi_tensor_idx_len = [item.e1_multi_tensor_idx.shape[0] for item in data_list]
        max_e1_multi_tensor_idx_len = max(e1_multi_tensor_idx_len)
        e1_multi_tensor_idx_pad = []
        for item in data_list:
            if item.e1_multi_tensor_idx.shape[0] < max_e1_multi_tensor_idx_len:
                need_pad_length = max_e1_multi_tensor_idx_len - item.e1_multi_tensor_idx.shape[0]
                pad = torch.zeros(need_pad_length).fill_(Vocab.PAD)
                e1_multi_tensor_idx_pad.append(torch.cat((item.e1_multi_tensor_idx, pad.long()), dim=0).unsqueeze(0))
            elif item.e1_multi_tensor_idx.shape[0] == max_e1_multi_tensor_idx_len:
                e1_multi_tensor_idx_pad.append(item.e1_multi_tensor_idx.unsqueeze(0))
            else:
                raise RuntimeError("Size mismatch error")

        e1_multi_tensor_idx = torch.cat(e1_multi_tensor_idx_pad, dim=0)

        e1_multi_len = [item.e1_multi_tensor.shape[0] for item in data_list]
        max_e1_multi_len = max(e1_multi_len)
        e1_multi_pad = []
        for item in data_list:
            if item.e1_multi_tensor.shape[0] < max_e1_multi_len:
                need_pad_length = max_e1_multi_len - item.e1_multi_tensor.shape[0]
                pad = torch.zeros(need_pad_length).fill_(Vocab.PAD)
                e1_multi_pad.append(torch.cat((item.e1_multi_tensor, pad.long()), dim=0).unsqueeze(0))
            elif item.e1_multi_tensor.shape[0] == max_e1_multi_len:
                e1_multi_pad.append(item.e1_multi_tensor.unsqueeze(0))
            else:
                raise RuntimeError("Size mismatch error")

        e1_multi = torch.cat(e1_multi_pad, dim=0)

        # do padding here
        e2_multi_tensor_idx_len = [item.e2_multi_tensor_idx.shape[0] for item in data_list]
        max_e2_multi_tensor_idx_len = max(e2_multi_tensor_idx_len)
        e2_multi_tensor_idx_pad = []
        for item in data_list:
            if item.e2_multi_tensor_idx.shape[0] < max_e2_multi_tensor_idx_len:
                need_pad_length = max_e2_multi_tensor_idx_len - item.e2_multi_tensor_idx.shape[0]
                pad = torch.zeros(need_pad_length).fill_(Vocab.PAD)
                e2_multi_tensor_idx_pad.append(torch.cat((item.e2_multi_tensor_idx, pad.long()), dim=0).unsqueeze(0))
            elif item.e2_multi_tensor_idx.shape[0] == max_e2_multi_tensor_idx_len:
                e2_multi_tensor_idx_pad.append(item.e2_multi_tensor_idx.unsqueeze(0))
            else:
                raise RuntimeError("Size mismatch error")

        e2_multi_tensor_idx = torch.cat(e2_multi_tensor_idx_pad, dim=0)

        e2_multi_len = [item.e2_multi_tensor.shape[0] for item in data_list]
        max_e2_multi_len = max(e2_multi_len)
        e2_multi_pad = []
        for item in data_list:
            if item.e2_multi_tensor.shape[0] < max_e2_multi_len:
                need_pad_length = max_e2_multi_len - item.e2_multi_tensor.shape[0]
                pad = torch.zeros(need_pad_length).fill_(Vocab.PAD)
                e2_multi_pad.append(torch.cat((item.e2_multi_tensor, pad.long()), dim=0).unsqueeze(0))
            elif item.e2_multi_tensor.shape[0] == max_e2_multi_len:
                e2_multi_pad.append(item.e2_multi_tensor.unsqueeze(0))
            else:
                raise RuntimeError("Size mismatch error")

        e2_multi = torch.cat(e2_multi_pad, dim=0)

        return {'e1': e1,
                'rel': rel,
                'e2_multi': e2_multi,
                'e2_multi_tensor_idx': e2_multi_tensor_idx,
                'e2': e2,
                'rel_eval': rel_eval,
                'e1_multi': e1_multi,
                'e1_multi_tensor_idx': e1_multi_tensor_idx}


class Text2LabelDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir, **kwargs):
        self.data_item_type = Text2LabelDataItem
        super(Text2LabelDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.

        For Text2LabelDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).

        Examples
        --------
        input: How far is it from Denver to Aspen ?    NUM

        DataItem: input_text="How far is it from Denver to Aspen ?", output_label="NUM"

        Parameters
        ----------
        file_path: str
            The path of the input file.

        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            for line in f:
                input, output = line.split('\t')
                data_item = Text2LabelDataItem(input_text=input.strip(), output_label=output.strip(),
                                               tokenizer=self.tokenizer)
                data.append(data_item)

        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=self.max_word_vocab_size,
                                       min_word_vocab_freq=self.min_word_vocab_freq,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=self.word_emb_size,
                                       share_vocab=True)
        self.vocab_model = vocab_model

        # label encoding
        all_labels = {item.output_label for item in self.train + self.test}
        if 'val' in self.__dict__:
            all_labels = all_labels.union({item.output_label for item in self.val})

        self.le = preprocessing.LabelEncoder()
        self.le.fit(list(all_labels))
        self.num_classes = len(self.le.classes_)

    def vectorization(self, data_items):
        for item in data_items:
            graph: GraphData = item.graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.vocab_model.in_word_vocab.getIndex(node_token)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            token_matrix = torch.LongTensor(token_matrix)
            graph.node_features['token_id'] = token_matrix

            item.output = self.le.transform([item.output_label])[0]

    @staticmethod
    def collate_fn(data_list: [Text2LabelDataItem]):
        graph_data = [deepcopy(item.graph) for item in data_list]

        tgt = [deepcopy(item.output) for item in data_list]
        tgt_tensor = torch.LongTensor(tgt)

        return [graph_data, tgt_tensor]


class DoubleText2TextDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir, share_vocab=True, **kwargs):
        self.data_item_type = DoubleText2TextDataItem
        self.share_vocab = share_vocab
        super(DoubleText2TextDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.

        For DoubleText2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).
        # TODO: update example

        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )

        DataItem: input_text="list job use languageid0", input_text2="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"

        Parameters
        ----------
        file_path: str
            The path of the input file.

        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            for line in f:
                input, input2, output = line.split('\t')
                data_item = DoubleText2TextDataItem(input_text=input.strip(), input_text2=input2.strip(),
                                                    output_text=output.strip(), tokenizer=self.tokenizer,
                                                    share_vocab=self.share_vocab)
                data.append(data_item)
        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=self.max_word_vocab_size,
                                       min_word_vocab_freq=self.min_word_vocab_freq,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=self.word_emb_size,
                                       share_vocab=self.share_vocab)
        self.vocab_model = vocab_model

        return self.vocab_model

    def vectorization(self, data_items):
        if self.topology_builder == IEBasedGraphConstruction:
            use_ie = True
        else:
            use_ie = False
        for item in data_items:
            graph: GraphData = item.graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.vocab_model.in_word_vocab.getIndex(node_token, use_ie)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            if self.topology_builder == IEBasedGraphConstruction:
                for i in range(len(token_matrix)):
                    token_matrix[i] = np.array(token_matrix[i][0])
                token_matrix = pad_2d_vals_no_size(token_matrix)
                token_matrix = torch.tensor(token_matrix, dtype=torch.long)
                graph.node_features['token_id'] = token_matrix
                pass
            else:
                token_matrix = torch.tensor(token_matrix, dtype=torch.long)
                graph.node_features['token_id'] = token_matrix

            if use_ie and 'token' in graph.edge_attributes[0].keys():
                edge_token_matrix = []
                for edge_idx in range(graph.get_edge_num()):
                    edge_token = graph.edge_attributes[edge_idx]['token']
                    edge_token_id = self.vocab_model.in_word_vocab.getIndex(edge_token, use_ie)
                    graph.edge_attributes[edge_idx]['token_id'] = edge_token_id
                    edge_token_matrix.append([edge_token_id])
                if self.topology_builder == IEBasedGraphConstruction:
                    for i in range(len(edge_token_matrix)):
                        edge_token_matrix[i] = np.array(edge_token_matrix[i][0])
                    edge_token_matrix = pad_2d_vals_no_size(edge_token_matrix)
                    edge_token_matrix = torch.tensor(edge_token_matrix, dtype=torch.long)
                    graph.edge_features['token_id'] = edge_token_matrix

            item.input_text2 = self.tokenizer(item.input_text2)
            input_token_id2 = self.vocab_model.in_word_vocab.to_index_sequence_for_list(item.input_text2)
            input_token_id2 = np.array(input_token_id2)
            item.input_np2 = input_token_id2

            if self.lower_case:
                item.output_text = item.output_text.lower()

            item.output_text = self.tokenizer(item.output_text)

            tgt_token_id = self.vocab_model.in_word_vocab.to_index_sequence_for_list(item.output_text)
            tgt_token_id.append(self.vocab_model.in_word_vocab.EOS)
            tgt_token_id = np.array(tgt_token_id)
            item.output_np = tgt_token_id
            item.output_text = ' '.join(item.output_text)

    @staticmethod
    def collate_fn(data_list: [Text2TextDataItem]):
        graph_data = []
        input_tensor2, input_length2, input_text2 = [], [], []
        tgt_tensor, tgt_text = [], []
        for item in data_list:
            graph_data.append(deepcopy(item.graph))
            input_tensor2.append(deepcopy(item.input_np2))
            input_length2.append(len(item.input_np2))
            input_text2.append(deepcopy(item.input_text2))
            tgt_tensor.append(deepcopy(item.output_np))
            tgt_text.append(deepcopy(item.output_text))

        input_tensor2 = torch.LongTensor(pad_2d_vals_no_size(input_tensor2))
        input_length2 = torch.LongTensor(input_length2)
        tgt_tensor = torch.LongTensor(pad_2d_vals_no_size(tgt_tensor))

        return {'graph_data': graph_data,
                'input_tensor2': input_tensor2,
                'input_text2': input_text2,
                'tgt_tensor': tgt_tensor,
                'tgt_text': tgt_text,
                'input_length2': input_length2}


class SequenceLabelingDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir, tag_types, **kwargs):
        self.data_item_type = SequenceLabelingDataItem
        self.tag_types = tag_types
        super(SequenceLabelingDataset, self).__init__(root_dir, topology_builder, topology_subdir, **kwargs)

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.
        For SequenceLabelingDataset, the format of the input file should contain lines of tokens, each line representing one
        record of token at first column and its tag at the last column. 
        Examples
        --------
        "EU       I-ORG "
         rejects  O
         German   I-MISC
         
        Parameters
        ----------

        """
        data = []
        input = []
        output = []
        with open(file_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if len(line) > 1 and line[0] != '-':
                    if line[0] != '.':
                        input.append(line.strip().split(' ')[0])
                        output.append(line.strip().split(' ')[-1])
                    if line[0] == '.':
                        input.append(line.strip().split(' ')[0])
                        output.append(line.strip().split(' ')[-1])
                        if len(input) >= 2:
                            data_item = SequenceLabelingDataItem(input_text=input, output_tags=output,
                                                                 tokenizer=self.tokenizer)
                            data.append(data_item)
                            input = []
                            output = []

        return data

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=None,
                                       min_word_vocab_freq=1,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=300,
                                       share_vocab=True)
        self.vocab_model = vocab_model

        return self.vocab_model

    def vectorization(self, data_items):
        for item in data_items:
            graph: GraphData = item.graph
            token_matrix = []
            for node_idx in range(graph.get_node_num()):
                node_token = graph.node_attributes[node_idx]['token']
                node_token_id = self.vocab_model.in_word_vocab.getIndex(node_token)
                graph.node_attributes[node_idx]['token_id'] = node_token_id
                token_matrix.append([node_token_id])
            token_matrix = torch.tensor(token_matrix, dtype=torch.long)
            graph.node_features['token_id'] = token_matrix

            tgt = item.output_tag
            tgt_tag_id = [self.tag_types.index(tgt_.strip()) for tgt_ in tgt]

            tgt_tag_id = torch.tensor(tgt_tag_id)
            item.output_id = tgt_tag_id

    @staticmethod
    def collate_fn(data_list: [SequenceLabelingDataItem]):
        tgt_tag = []
        graph_data = []
        for item in data_list:
            # if len(item.graph.node_attributes)== len(item.output_id):
            graph_data.append(deepcopy(item.graph))
            tgt_tag.append(deepcopy(item.output_id))

        # tgt_tags = torch.cat(tgt_tag, dim=0)
        #return [graph_data, tgt_tag]
        return {"graph_data": graph_data,
            "tgt_tag": tgt_tag}
        

class CNNSeq2SeqDataset(Dataset):
    def __init__(self, root_dir, topology_builder, topology_subdir,
                 share_vocab=True, word_emb_size=300,
                 dynamic_graph_type=None,
                 dynamic_init_topology_builder=None,
                 dynamic_init_topology_aux_args=None
                 ):
        self.data_item_type = Text2TextDataItem_seq2seq
        self.share_vocab = share_vocab
        self.tokenizer = None
        super(CNNSeq2SeqDataset, self).__init__(root=root_dir,
                                                tokenizer=self.tokenizer,
                                                topology_builder=topology_builder,
                                                topology_subdir=topology_subdir,
                                                 share_vocab=share_vocab, word_emb_size=word_emb_size,
                                                 dynamic_graph_type=dynamic_graph_type,
                                                 dynamic_init_topology_builder=dynamic_init_topology_builder,
                                                 dynamic_init_topology_aux_args=dynamic_init_topology_aux_args)


    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train_3.json', 'val': "val_3.json", 'test': 'test_3.json'}
        # return {'train': 'train-0.json', 'val': "val-0.json", 'test': 'test-0.json'}
        # return {'train': 'train.json', 'val': "val.json", 'test': 'test.json'}

    @property
    def processed_file_names(self):
        """At least 3 reserved keys should be fiiled: 'vocab', 'data' and 'split_ids'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        # raise NotImplementedError("It shouldn't be called now")
        return

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=50000,
                                       min_word_vocab_freq=8,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=self.word_emb_size,
                                       share_vocab=self.share_vocab)
        self.vocab_model = vocab_model

        return self.vocab_model

    def _process(self):
        if all([os.path.exists(processed_path) for processed_path in self.processed_file_paths.values()]):
            if 'val_split_ratio' in self.__dict__:
                UserWarning(
                    "Loading existing processed files on disk. Your `val_split_ratio` might not work since the data have"
                    "already been split.")
            return

        os.makedirs(self.processed_dir, exist_ok=True)

        self.read_raw_data()

        self.build_vocab()

        self.vectorization(self.train)
        self.vectorization(self.test)
        if 'val' in self.__dict__:
            self.vectorization(self.val)

        data_to_save = {'train': self.train, 'test': self.test}
        if 'val' in self.__dict__:
            data_to_save['val'] = self.val
        torch.save(data_to_save, self.processed_file_paths['data'])

    def parse_file(self, file_path):
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.
        For Text2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).
        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )
        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"
        Parameters
        ----------
        file_path: str
            The path of the input file.
        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            examples = json.load(f)
            for example_dict in examples:
                input = ' '.join(' '.join(example_dict['article']).split()[:500])
                output = ' '.join(' '.join(['<t> '+sent[0]+' . </t>' for sent in example_dict['highlight']]).split()[:99])
                if input=='' or output=='':
                    continue
                data_item = Text2TextDataItem_seq2seq(input_text=input,
                                                      output_text=output,
                                                      tokenizer=self.tokenizer,
                                                      share_vocab=self.share_vocab)
                data.append(data_item)
        return data

    def vectorization(self, data_items):
        for item in data_items:
            src = item.input_text
            src_token_id = self.vocab_model.in_word_vocab.to_index_sequence(src)
            src_token_id.append(self.vocab_model.in_word_vocab.EOS)
            src_token_id = np.array(src_token_id)
            item.input_np = src_token_id

            tgt = item.output_text
            tgt_token_id = self.vocab_model.out_word_vocab.to_index_sequence(tgt)
            tgt_token_id.append(self.vocab_model.in_word_vocab.EOS)
            tgt_token_id = np.array(tgt_token_id)
            item.output_np = tgt_token_id

    @staticmethod
    def collate_fn(data_list: [Text2TextDataItem_seq2seq]):
        input_numpy = [deepcopy(item.input_np) for item in data_list]
        src_len = [item.input_np.shape[0] for item in data_list]
        input_pad = pad_2d_vals_no_size(input_numpy)

        src_seq = torch.from_numpy(input_pad).long()
        src_len = torch.LongTensor(src_len)

        output_numpy = [deepcopy(item.output_np) for item in data_list]
        output_pad = pad_2d_vals_no_size(output_numpy)

        tgt_seq = torch.from_numpy(output_pad).long()

        src_str = [deepcopy(item.input_text) for item in data_list]
        tgt_str = [deepcopy(item.output_text) for item in data_list]

        sorted_x_len, indx = torch.sort(src_len, 0, descending=True)
        src_seq = src_seq[indx]
        tgt_seq = tgt_seq[indx]
        src_str = np.array(src_str)[indx]
        tgt_str = np.array(tgt_str)[indx]

        return [src_seq, sorted_x_len, tgt_seq, src_str, tgt_str]


import os
import argparse
import numpy as np
from scipy import sparse
import networkx as nx
import time
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.backends.cudnn as cudnn
import dgl
from dgl import DGLGraph
import dgl.function as fn
from dgl.data import register_data_args, load_data

from ..utils import EarlyStopping
from ...modules.graph_construction import NodeEmbeddingBasedGraphConstruction, NodeEmbeddingBasedRefinedGraphConstruction
from ...modules.utils.vocab_utils import VocabModel
from ...modules.utils.padding_utils import pad_2d_vals_no_size
from ...modules.utils.generic_utils import to_cuda


def accuracy(logits, labels):
    _, indices = torch.max(logits, dim=1)
    correct = torch.sum(indices == labels)
    return correct.item() * 1.0 / len(labels)

def evaluate(model, g, labels, mask):
    model.eval()
    with torch.no_grad():
        logits, _ = model(g)
        logits = logits[mask]
        labels = labels[mask]
        return accuracy(logits, labels)

def prepare_dgl_graph_data(args):
    data = load_data(args)
    features = torch.FloatTensor(data.features)
    labels = torch.LongTensor(data.labels)
    if hasattr(torch, 'BoolTensor'):
        train_mask = torch.BoolTensor(data.train_mask)
        val_mask = torch.BoolTensor(data.val_mask)
        test_mask = torch.BoolTensor(data.test_mask)
    else:
        train_mask = torch.ByteTensor(data.train_mask)
        val_mask = torch.ByteTensor(data.val_mask)
        test_mask = torch.ByteTensor(data.test_mask)

    num_feats = features.shape[1]
    n_classes = data.num_labels
    n_edges = data.graph.number_of_edges()
    print("""----Data statistics------'
      #Edges %d
      #Classes %d
      #Train samples %d
      #Val samples %d
      #Test samples %d""" %
          (n_edges, n_classes,
           train_mask.int().sum().item(),
           val_mask.int().sum().item(),
           test_mask.int().sum().item()))

    g = data.graph
    # add self loop
    g.remove_edges_from(nx.selfloop_edges(g))
    g = DGLGraph(g)
    g.add_edges(g.nodes(), g.nodes())
    n_edges = g.number_of_edges()

    data = {'features': features,
            'graph': g,
            'train_mask': train_mask,
            'val_mask': val_mask,
            'test_mask': test_mask,
            'labels': labels,
            'num_feats': num_feats,
            'n_classes': n_classes,
            'n_edges': n_edges}

    return data

def prepare_ogbn_graph_data(args):
    from ogb.nodeproppred import DglNodePropPredDataset

    dataset = DglNodePropPredDataset(name=args.dataset)

    split_idx = dataset.get_idx_split()
    train_idx, val_idx, test_idx = torch.LongTensor(split_idx['train']), torch.LongTensor(split_idx['valid']), torch.LongTensor(split_idx['test'])
    g, labels = dataset[0] # graph: dgl graph object, label: torch tensor of shape (num_nodes, num_tasks)
    features = torch.Tensor(g.ndata['feat'])
    labels = torch.LongTensor(labels).squeeze(-1)

    # add self loop
    # no duplicate self loop will be added for nodes already having self loops
    new_g = dgl.transform.add_self_loop(g)


    # edge_index = data[0]['edge_index']
    # adj = to_undirected(edge_index, num_nodes=data[0]['num_nodes'])
    # assert adj.diagonal().sum() == 0 and adj.max() <= 1 and (adj != adj.transpose()).sum() == 0

    num_feats = features.shape[1]
    n_classes = labels.max().item() + 1
    n_edges = new_g.number_of_edges()
    print("""----Data statistics------'
      #Edges %d
      #Classes %d
      #Train samples %d
      #Val samples %d
      #Test samples %d""" %
          (n_edges, n_classes,
           train_idx.shape[0],
           val_idx.shape[0],
           test_idx.shape[0]))

    data = {'features': features,
            'graph': new_g,
            'train_mask': train_idx,
            'val_mask': val_idx,
            'test_mask': test_idx,
            'labels': labels,
            'num_feats': num_feats,
            'n_classes': n_classes,
            'n_edges': n_edges}

    return data

def normalize_sparse_adj(mx):
    """Row-normalize sparse matrix: symmetric normalized Laplacian"""
    rowsum = np.array(mx.sum(1))
    r_inv_sqrt = np.power(rowsum, -0.5).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    r_mat_inv_sqrt = sparse.diags(r_inv_sqrt)

    return mx.dot(r_mat_inv_sqrt).transpose().dot(r_mat_inv_sqrt)

def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)

    return torch.sparse.FloatTensor(indices, values, shape)

def get_normalized_init_adj(graph):
    adj = graph.adjacency_matrix_scipy(return_edge_ids=False)
    adj = normalize_sparse_adj(adj)
    adj = sparse_mx_to_torch_sparse_tensor(adj)

    return adj

class GNNLayer(nn.Module):
    def __init__(self, input_size, output_size):
        super(GNNLayer, self).__init__()
        self.linear_out = nn.Linear(input_size, output_size, bias=False)

    def forward(self, dgl_graph, node_emb):
        with dgl_graph.local_scope():
            dgl_graph.srcdata.update({'ft': node_emb})
            dgl_graph.update_all(fn.u_mul_e('ft', 'edge_weight', 'm'),
                                     fn.sum('m', 'ft'))
            agg_vec = dgl_graph.dstdata['ft']
            new_node_vec = self.linear_out(agg_vec)

            return new_node_vec

class GNNClassifier(nn.Module):
    def __init__(self,
                input_size,
                hidden_size,
                output_size,
                feat_drop=0,
                layer_drop=0.5,
                activation=F.relu):
        super(GNNClassifier, self).__init__()
        self.feat_drop = feat_drop
        self.layer_drop = layer_drop
        self.activation = activation
        self.gnn_layer1 = GNNLayer(input_size, hidden_size)
        self.gnn_layer2 = GNNLayer(hidden_size, output_size)


    def forward(self, graph):
        node_feat = graph.ndata['node_feat']
        node_feat = F.dropout(node_feat, self.feat_drop, training=self.training)

        node_emb = self.activation(self.gnn_layer1(graph, node_feat))
        node_emb = F.dropout(node_emb, self.layer_drop, training=self.training)

        logits = self.gnn_layer2(graph, node_emb)

        return logits

class DynamicGNNClassifier(nn.Module):
    def __init__(self,
                word_vocab,
                embedding_styles,
                input_size,
                hidden_size,
                output_size,
                gl_type,
                gl_metric_type='weighted_cosine',
                gl_num_heads=1,
                gl_top_k=None,
                gl_epsilon=None,
                gl_smoothness_ratio=None,
                gl_connectivity_ratio=None,
                gl_sparsity_ratio=None,
                gl_input_size=None,
                gl_hidden_size=None,
                init_adj_alpha=None,
                feat_drop=0,
                layer_drop=0.5,
                activation=F.relu,
                device=None):
        super(DynamicGNNClassifier, self).__init__()
        self.gl_type = gl_type
        if self.gl_type == 'node_emb':
            self.graph_learner = NodeEmbeddingBasedGraphConstruction(
                                                        word_vocab,
                                                        embedding_styles,
                                                        sim_metric_type=gl_metric_type,
                                                        num_heads=gl_num_heads,
                                                        top_k_neigh=gl_top_k,
                                                        epsilon_neigh=gl_epsilon,
                                                        smoothness_ratio=gl_smoothness_ratio,
                                                        connectivity_ratio=gl_connectivity_ratio,
                                                        sparsity_ratio=gl_sparsity_ratio,
                                                        input_size=gl_input_size,
                                                        hidden_size=gl_hidden_size,
                                                        fix_word_emb=True,
                                                        dropout=None,
                                                        device=device)

        elif self.gl_type == 'node_edge_emb':
            raise NotImplementedError()
        elif self.gl_type == 'node_emb_refined':
            self.graph_learner = NodeEmbeddingBasedRefinedGraphConstruction(
                                                        word_vocab,
                                                        embedding_styles,
                                                        init_adj_alpha,
                                                        sim_metric_type=gl_metric_type,
                                                        num_heads=gl_num_heads,
                                                        top_k_neigh=gl_top_k,
                                                        epsilon_neigh=gl_epsilon,
                                                        smoothness_ratio=gl_smoothness_ratio,
                                                        connectivity_ratio=gl_connectivity_ratio,
                                                        sparsity_ratio=gl_sparsity_ratio,
                                                        input_size=gl_input_size,
                                                        hidden_size=gl_hidden_size,
                                                        fix_word_emb=True,
                                                        dropout=None,
                                                        device=device)
        else:
            raise RuntimeError('Unknown gl_type: {}'.format(self.gl_type))

        self.gnn_clf = GNNClassifier(
                                input_size,
                                hidden_size,
                                output_size,
                                feat_drop=feat_drop,
                                layer_drop=layer_drop,
                                activation=activation)

    def forward(self, graph):
        node_feat = graph.ndata['node_feat']

        if self.gl_type == 'node_emb_refined':
            new_graph = self.graph_learner.topology(node_feat, graph.init_adj, node_mask=None)
        else:
            new_graph = self.graph_learner.topology(node_feat, node_mask=None)

        # convert GraphData to DGLGraph
        dgl_graph = new_graph.to_dgl()
        dgl_graph.ndata['node_feat'] = node_feat
        dgl_graph.edata['edge_weight'] = new_graph.edge_features['edge_weight']
        dgl_graph.graph_reg = new_graph.graph_attributes['graph_reg']
        logits = self.gnn_clf(dgl_graph)

        return logits, dgl_graph

def main(args, seed):
    # Configure
    np.random.seed(seed)
    torch.manual_seed(seed)

    if not args.no_cuda and torch.cuda.is_available():
        print('[ Using CUDA ]')
        device = torch.device('cuda' if args.gpu < 0 else 'cuda:%d' % args.gpu)
        cudnn.benchmark = True
        torch.cuda.manual_seed(seed)
    else:
        device = torch.device('cpu')


    # load and preprocess dataset
    if args.dataset.startswith('ogbn'):
        # Open Graph Benchmark datasets
        data = prepare_ogbn_graph_data(args)
    else:
        # DGL datasets
        data = prepare_dgl_graph_data(args)

    features, g, train_mask, val_mask, test_mask, labels, num_feats, n_classes, n_edges\
                             = data['features'], data['graph'], data['train_mask'], \
                             data['val_mask'], data['test_mask'], data['labels'], \
                             data['num_feats'], data['n_classes'], data['n_edges']

    features = features.to(device)
    labels = labels.to(device)
    train_mask = train_mask.to(device)
    val_mask = val_mask.to(device)
    test_mask = test_mask.to(device)
    g.ndata['node_feat'] = features

    if args.gl_type == 'node_emb_refined':
        init_adj = get_normalized_init_adj(g).to(device)
        g.init_adj = init_adj

    raw_text_data = [['I like nlp.', 'Same here!'], ['I like graph.', 'Same here!']]
    vocab_model = VocabModel(raw_text_data, max_word_vocab_size=None,
                                min_word_vocab_freq=1,
                                word_emb_size=300)

    src_text_seq = list(zip(*raw_text_data))[0]
    src_idx_seq = [vocab_model.word_vocab.to_index_sequence(each) for each in src_text_seq]
    src_len = torch.LongTensor([len(each) for each in src_idx_seq])
    num_seq = torch.LongTensor([len(src_len)])
    input_tensor = torch.LongTensor(pad_2d_vals_no_size(src_idx_seq))




    embedding_styles = {'word_emb_type': 'w2v',
                        'node_edge_emb_strategy': 'bilstm',
                        'seq_info_encode_strategy': 'none'}

    # create model
    model = DynamicGNNClassifier(
                    vocab_model.word_vocab,
                    embedding_styles,
                    num_feats,
                    args.num_hidden,
                    n_classes,
                    args.gl_type,
                    gl_metric_type=args.gl_metric_type,
                    gl_num_heads=args.gl_num_heads,
                    gl_top_k=args.gl_top_k,
                    gl_epsilon=args.gl_epsilon,
                    gl_smoothness_ratio=args.gl_smoothness_ratio,
                    gl_connectivity_ratio=args.gl_connectivity_ratio,
                    gl_sparsity_ratio=args.gl_sparsity_ratio,
                    gl_input_size=num_feats,
                    gl_hidden_size=args.gl_num_hidden,
                    init_adj_alpha=args.init_adj_alpha,
                    feat_drop=args.in_drop,
                    layer_drop=args.layer_drop,
                    activation=F.relu,
                    device=device)


    print(model)
    model.to(device)

    if args.early_stop:
        stopper = EarlyStopping('{}.{}'.format(args.save_model_path, seed), patience=args.patience)

    loss_fcn = torch.nn.CrossEntropyLoss()

    # use optimizer
    optimizer = torch.optim.Adam(
        model.parameters(), lr=args.lr, weight_decay=args.weight_decay)

    # initialize graph
    dur = []
    for epoch in range(args.epochs):
        model.train()
        if epoch >= 3:
            t0 = time.time()
        # forward
        logits, dgl_graph = model(g)
        loss = loss_fcn(logits[train_mask], labels[train_mask])
        loss += dgl_graph.graph_reg

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if epoch >= 3:
            dur.append(time.time() - t0)

        train_acc = accuracy(logits[train_mask], labels[train_mask])

        if args.fastmode:
            val_acc = accuracy(logits[val_mask], labels[val_mask])
        else:
            val_acc = evaluate(model, g, labels, val_mask)
            if args.early_stop:
                if stopper.step(val_acc, model):
                    break

        print("Epoch {:05d} | Time(s) {:.4f} | Loss {:.4f} | TrainAcc {:.4f} |"
              " ValAcc {:.4f} | ETputs(KTEPS) {:.2f}".
              format(epoch, np.mean(dur), loss.item(), train_acc,
                     val_acc, n_edges / np.mean(dur) / 1000))

    print()
    if args.early_stop:
        model = stopper.load_checkpoint(model)
        print('Restored best saved model')
        os.remove(stopper.save_model_path)
        print('Removed best saved model file to save disk space')

    acc = evaluate(model, g, labels, test_mask)
    print("Test Accuracy {:.4f}".format(acc))

    return acc


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='DynamicGraphConstruction')
    register_data_args(parser)
    parser.add_argument("--num-runs", type=int, default=5,
                        help="number of runs")
    parser.add_argument("--no-cuda", action="store_true", default=False,
                        help="use CPU")
    parser.add_argument("--gpu", type=int, default=-1,
                        help="which GPU to use.")
    parser.add_argument("--epochs", type=int, default=1000,
                        help="number of training epochs")
    parser.add_argument("--gl-num-hidden", type=int, default=16,
                        help="number of hidden units for dynamic graph construction")
    parser.add_argument("--gl-top-k", type=int,
                        help="top k for graph sparsification")
    parser.add_argument("--gl-epsilon", type=float,
                        help="epsilon for graph sparsification")
    parser.add_argument("--gl-smoothness-ratio", type=float,
                        help="smoothness ratio for graph regularization loss")
    parser.add_argument("--gl-connectivity-ratio", type=float,
                        help="connectivity ratio for graph regularization loss")
    parser.add_argument("--gl-sparsity-ratio", type=float,
                        help="sparsity ratio for graph regularization loss")
    parser.add_argument("--gl-num-heads", type=int, default=1,
                        help="num of heads for dynamic graph construction")
    parser.add_argument("--gl-type", type=str, default='node_emb',
                        help=r"dynamic graph construction algorithm type, \
                        {'node_emb', 'node_edge_emb' and 'node_emb_refined'},\
                        default: 'node_emb'")
    parser.add_argument("--gl-metric-type", type=str, default='weighted_cosine',
                        help=r"similarity metric type for dynamic graph construction")
    parser.add_argument("--init-adj-alpha", type=float, default=0.8,
                        help="alpha ratio for combining initial graph adjacency matrix")
    parser.add_argument("--num-hidden", type=int, default=16,
                        help="number of hidden units")
    parser.add_argument("--in-drop", type=float, default=0,
                        help="input feature dropout")
    parser.add_argument("--layer-drop", type=float, default=0.5,
                        help="layer dropout")
    parser.add_argument("--lr", type=float, default=0.01,
                        help="learning rate")
    parser.add_argument('--weight-decay', type=float, default=5e-4,
                        help="weight decay")
    parser.add_argument('--early-stop', action='store_true', default=False,
                        help="indicates whether to use early stop or not")
    parser.add_argument("--patience", type=int, default=100,
                        help="early stopping patience")
    parser.add_argument('--fastmode', action="store_true", default=False,
                        help="skip re-evaluate the validation set")
    parser.add_argument('--save-model-path', type=str, default="ckpt",
                        help="path to the best saved model")
    args = parser.parse_args()
    args.save_model_path = ('{}_{}_gl_type_{}_gl_metric_type_{}_gl_heads_{}'
                            '_gl_topk_{}_gl_epsilon_{}_smoothness_{}_connectivity_{}'
                            '_sparsity_{}_init_adj_alpha_{}').format(
                                                    args.save_model_path,
                                                    args.dataset,
                                                    args.gl_type,
                                                    args.gl_metric_type,
                                                    args.gl_num_heads,
                                                    args.gl_top_k,
                                                    args.gl_epsilon,
                                                    args.gl_smoothness_ratio,
                                                    args.gl_connectivity_ratio,
                                                    args.gl_sparsity_ratio,
                                                    args.init_adj_alpha)
    print(args)

    np.random.seed(123)
    scores = []
    for _ in range(args.num_runs):
        seed = np.random.randint(10000)
        scores.append(main(args, seed))

    print("\nTest Accuracy ({} runs): mean {:.4f}, std {:.4f}".format(args.num_runs, np.mean(scores), np.std(scores)))



from graph4nlp.pytorch.modules.evaluation.cider import CIDEr

if __name__ == "__main__":
    import json

    scorer = CIDEr(df="corpus")
    pred_file_path = "/home/shiina/shiina/question/iq/pred.json"
    gt_file_path = "/home/shiina/shiina/question/iq/gt.json"
    with open(gt_file_path, "r") as f:
        gt = json.load(f)
        print(gt[0])
        gts = []
        for i in gt:
            for j in i:
                gts.append(str(j))
    with open(pred_file_path, "r") as f:
        pred = json.load(f)
        print(pred[1])
        preds = []
        for i in pred:
            for j in i:
                preds.append(str(j))
    print(len(gts), len(preds))
    score, scores = scorer.calculate_scores(gts, preds)
    print(score)
    print(len(scores))
